using System;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.VR;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Blazedust.CUAM
{

    // Based on ResetVROrientation on how to add buttons to the menu-canvas.
    public class CanvasButtons
    {
        
        public static Canvas canvas = null;

        public bool forceEnabled;

        private bool isDesktopMode = false;

        private CustomUnityAssetManager core;

        private UIDynamicButton btnAppend;
        private UIDynamicButton btnReplace;
        private UIDynamicButton btnRemoveLast;
        private UIDynamicButton btnRemoveAll;

        private List<Transform> createdButtons;

        // private bool buttonsEnabled = false;

        private bool lastButtonsOverallEnabled = false;
        private bool lastButtonsVAMPresetsEnabled = false;
        private bool lastButtonsVAMMergePresetsEnabled = false;
        private bool lastButtonsVAMPluginPresetButton = false;
        private bool lastButtonsCUAMEnabled = false;
        private bool lastButtonsCUAMCompactLayetEnabled = false;
        private bool lastButtonsCUAMToggleTriggersEnabled = false;
        private string lastButtonsLayout = "";
        private float lastButtonsXPosOffset = 0f;
        private float lastButtonsYPosOffset = 0f;

        private string lastButtonsVRLayout = "";
        private float lastButtonsVRXPosOffset = 0f;
        private float lastButtonsVRYPosOffset = 0f;
        private float lastButtonsVRZPosOffset = 0f;

        private float lastButtonsVRXRotOffset = 0f;
        private float lastButtonsVRYRotOffset = 0f;

        private bool lastButtonStateEnabled = false;

        private int lastRemoveAllLinkedAssetQuickCounter = 0;

        public CanvasButtons(CustomUnityAssetManager c)
        {
            core = c;
            createdButtons = new List<Transform>();
        }

        private bool CheckButtonsLayoutChanged()
        {
            Atom selectedAtom = SuperController.singleton.GetSelectedAtom();
            bool buttonsOverallEnabled = core.LogicEnabled && (selectedAtom != null && selectedAtom.type == "Person" || forceEnabled && SuperController.singleton.MainHUDVisible && SuperController.singleton.activeUI == SuperController.ActiveUI.MainMenu);
            //public JSONStorableBool quickCanvasVAMPresetsButtons;
            //public JSONStorableBool quickCanvasCUAMButtons;
            //public JSONStorableStringChooser quickCanvasButtonsType;
            //public JSONStorableFloat quickCanvasButtonsXOffset;
            //public JSONStorableFloat quickCanvasButtonsYOffset;
            bool layoutChanged = lastButtonsOverallEnabled != buttonsOverallEnabled ||
                                 lastButtonsVAMPresetsEnabled != core.options.quickCanvasVAMPresetButtons.val ||
                                 lastButtonsVAMPluginPresetButton != core.options.quickCanvasVAMPluginPresetButton.val ||
                                 lastButtonsVAMMergePresetsEnabled != core.options.quickCanvasVAMMergePresetsButtons.val ||
                                 lastButtonsCUAMEnabled != core.options.quickCanvasCUAMButtons.val ||
                                 lastButtonsCUAMCompactLayetEnabled != core.options.quickCanvasCUAMButtonsCompactLayout.val ||
                                 lastButtonsCUAMToggleTriggersEnabled != core.options.quickCanvasCUAMToggleTriggers.val ||
                                 lastButtonsLayout != core.options.quickCanvasButtonsType.val ||
                                 lastButtonsXPosOffset != core.options.quickCanvasButtonsXOffset.val ||
                                 lastButtonsYPosOffset != core.options.quickCanvasButtonsYOffset.val ||
                                 lastButtonsVRLayout != core.options.quickCanvasButtonsVRType.val ||
                                 lastButtonsVRXPosOffset != core.options.quickCanvasButtonsVRXOffset.val ||
                                 lastButtonsVRYPosOffset != core.options.quickCanvasButtonsVRYOffset.val ||
                                 lastButtonsVRZPosOffset != core.options.quickCanvasButtonsVRZOffset.val ||
                                 lastButtonsVRXRotOffset != core.options.quickCanvasButtonsVRXRotOffset.val ||
                                 lastButtonsVRYRotOffset != core.options.quickCanvasButtonsVRYRotOffset.val;

            if (!layoutChanged)
                return false;

            lastButtonsOverallEnabled = buttonsOverallEnabled;
            
            lastButtonsVAMPresetsEnabled = core.options.quickCanvasVAMPresetButtons.val;
            lastButtonsVAMPluginPresetButton = core.options.quickCanvasVAMPluginPresetButton.val;
            lastButtonsVAMMergePresetsEnabled = core.options.quickCanvasVAMMergePresetsButtons.val;
            lastButtonsCUAMEnabled = core.options.quickCanvasCUAMButtons.val;
            lastButtonsCUAMCompactLayetEnabled = core.options.quickCanvasCUAMButtonsCompactLayout.val;
            lastButtonsCUAMToggleTriggersEnabled = core.options.quickCanvasCUAMToggleTriggers.val;

            lastButtonsLayout = core.options.quickCanvasButtonsType.val;
            lastButtonsXPosOffset = core.options.quickCanvasButtonsXOffset.val;
            lastButtonsYPosOffset = core.options.quickCanvasButtonsYOffset.val;

            lastButtonsVRLayout = core.options.quickCanvasButtonsVRType.val;
            lastButtonsVRXPosOffset = core.options.quickCanvasButtonsVRXOffset.val;
            lastButtonsVRYPosOffset = core.options.quickCanvasButtonsVRYOffset.val;
            lastButtonsVRZPosOffset = core.options.quickCanvasButtonsVRZOffset.val;
            lastButtonsVRXRotOffset = core.options.quickCanvasButtonsVRXRotOffset.val;
            lastButtonsVRYRotOffset = core.options.quickCanvasButtonsVRYRotOffset.val;

            return true;
        }

        public void UpdateButtonLayout()
        {
            if (CheckButtonsLayoutChanged())
            {
                UpdateButtons();
            }

            // check loading/non-loading and button texts
            if (lastButtonsCUAMEnabled)
            {
                bool buttonStateEnabled = (!SuperController.singleton.isLoading && !core.loading.WaitingForPostDelay());
                if (lastButtonStateEnabled != buttonStateEnabled)
                {
                    if (buttonStateEnabled)
                    {
                        if (btnAppend != null)
                            btnAppend.buttonColor = new Color(0.75f, 0.95f, 0.60f);

                        if (btnReplace != null)
                            btnReplace.buttonColor = new Color(0.5f, 0.9f, 0.9f);

                        if (btnRemoveLast != null)
                            btnRemoveLast.buttonColor = new Color(1.0f, 0.7f, 1.0f);

                        if (btnRemoveAll != null)
                            btnRemoveAll.buttonColor = new Color(1.0f, 0.7f, 0.7f);
                    }
                    else
                    {
                        if (btnAppend != null)
                            btnAppend.buttonColor = new Color(0.6f, 0.5f, 0.5f);

                        if (btnReplace != null)
                            btnReplace.buttonColor = new Color(0.6f, 0.5f, 0.5f);

                        if (btnRemoveLast != null)
                            btnRemoveLast.buttonColor = new Color(0.6f, 0.5f, 0.5f);

                        if (btnRemoveAll != null)
                            btnRemoveAll.buttonColor = new Color(0.6f, 0.5f, 0.5f);
                    }
                    lastButtonStateEnabled = buttonStateEnabled;
                }

                if (lastRemoveAllLinkedAssetQuickCounter != core.removeAllLinkedAssetQuickCounter)
                {
                    if (btnRemoveAll != null)
                    {
                        if (core.removeAllLinkedAssetQuickCounter == 0)
                        {
                            if (lastButtonsCUAMCompactLayetEnabled)
                            {
                                btnRemoveAll.label = "All)"; 
                            } 
                            else
                            {
                                btnRemoveAll.label = "[CUAM]\nRemove All";
                            }
                        }
                        else
                        {
                            if (lastButtonsCUAMCompactLayetEnabled)
                            {
                                btnRemoveAll.label = "(" + (CustomUnityAssetManager.FORCE_REMOVE_CLICKS - core.removeAllLinkedAssetQuickCounter).ToString() + ")";
                            }
                            else
                            {
                                btnRemoveAll.label = "Forced (" + (CustomUnityAssetManager.FORCE_REMOVE_CLICKS - core.removeAllLinkedAssetQuickCounter).ToString() + ")";
                            }
                        }
                    }
                    lastRemoveAllLinkedAssetQuickCounter = core.removeAllLinkedAssetQuickCounter;
                }
            }
        }

        private bool CheckPersonAtomSelectedForCUAMAction()
        {
            if (!lastButtonStateEnabled)
            {
                return false;
            }

            Atom selectedAtom = SuperController.singleton.GetSelectedAtom();
            if (selectedAtom != null && selectedAtom.type == "Person")
            {
                if (core.rootAtom == null || core.rootAtom.uid != selectedAtom.uid)
                {
                    core.SelectRootAtom(selectedAtom.uid);
                }
                return true;
            }
            return false;
        }

        private class ButtonLayoutInfo
        {
            public bool enabled;
            public bool isMergeButton;
            public float x;
            public float y;
            public float width;
            public string text;
            public Color buttonColor;
            public Color textColor;
            public Action callback;
            public Action<UIDynamicButton> createdDynamicButton;
            
            public ButtonLayoutInfo(bool enabledValue, bool mergeButton, string t, Color textC, Color buttonC, Action actionCallback, Action<UIDynamicButton> createdButton = null)
            {
                enabled = enabledValue;
                isMergeButton = mergeButton;
                x = 0f;
                y = 0f;
                if (isMergeButton)
                    width = -60f;
                else
                    width = 12f;
                text = t;
                buttonColor = buttonC;
                textColor = textC;
                callback = actionCallback;
                createdDynamicButton = createdButton;
            }

            public ButtonLayoutInfo(bool enabledValue, float posx, float posy)
            {
                enabled = enabledValue;
                x = posx;
                y = posy;
                width = 12f;
                text = "";
            }

            public ButtonLayoutInfo(bool enabledValue, float posx, float posy, string t = "")
            {
                enabled = enabledValue;
                x = posx;
                y = posy;
                width = 12f;
                text = t;
            }

            public ButtonLayoutInfo(bool enabledValue, float posx, float posy, float w = 12f, string t = "")
            {
                enabled = enabledValue;
                x = posx;
                y = posy;
                width = w;
                text = t;
            }
        }

        public void UpdateButtons()
        {
            // CreateButtons also clears previously buttons to be able to recreate them.
            CreateButtons(() => {

                btnAppend = null;
                btnReplace = null;
                btnRemoveLast = null;
                btnRemoveAll = null;

                if (!lastButtonsOverallEnabled)
                    return;

                // if core.options.enableCanvasButtons ...
                const float bottomY = -0.243f + 0.065f;
                const float leftX = -0.517f;
                const float buttonHeight = 0.045f;
                const float buttonWidth = 0.115f;

                float posx = leftX + (isDesktopMode ? lastButtonsXPosOffset : lastButtonsVRXPosOffset);
                float posy = bottomY + (isDesktopMode ? lastButtonsYPosOffset : lastButtonsVRYPosOffset);
                float posz = (isDesktopMode ? 0f : -0.01f + lastButtonsVRZPosOffset);

                List<ButtonLayoutInfo> buttons = new List<ButtonLayoutInfo>();
                // A merge button will collapse into the previously button
                // A disable button will be skipped
                // position is set automatically at the end based on layout + enabled + merge status of the button

                bool mergeClothingEnabled = lastButtonsVAMMergePresetsEnabled || true; // merge clothing always
                buttons.Add(new ButtonLayoutInfo(lastButtonsVAMPresetsEnabled, false, (mergeClothingEnabled ? "Select       \nClothing       " : "Select\nClothing"), Color.white, new Color((152f / 255f), (12f / 255f), (99f / 255f)), () => {
                    core.QuickLoadClothingPreset(SuperController.singleton.GetSelectedAtom());
                }));
                buttons.Add(new ButtonLayoutInfo(lastButtonsVAMPresetsEnabled && mergeClothingEnabled, true, "(M)", Color.white, new Color((222f / 255f), (112f / 255f), (129f / 255f)), () => {
                    core.QuickLoadClothingPreset(SuperController.singleton.GetSelectedAtom(), true);
                }));
                buttons.Add(new ButtonLayoutInfo(lastButtonsVAMPresetsEnabled, false, "Select\nAppearance", Color.white, new Color((40f / 255f), (95f / 255f), (95f / 255f)), () => {
                    core.QuickLoadAppearancePreset(SuperController.singleton.GetSelectedAtom());
                }));

                // CUAManager is intended to always merge plugins, to not ruin/destroy scenes when missclicking
                //buttons.Add(new ButtonLayoutInfo(lastButtonsVAMPluginPresetsButtons, false, "Merge\nPlugin", Color.white, new Color((126f / 255f), (12f / 255f), (152f / 255f)), () => {
                //    core.QuickLoadPluginPreset(SuperController.singleton.GetSelectedAtom(), true);
                //}));

                buttons.Add(new ButtonLayoutInfo(lastButtonsVAMPluginPresetButton, false, "Merge        \nPlugin       ", Color.white, new Color((126f / 255f), (12f / 255f), (152f / 255f)), () => {
                    core.QuickLoadPluginPreset(SuperController.singleton.GetSelectedAtom(), true);
                }));
                buttons.Add(new ButtonLayoutInfo(lastButtonsVAMPluginPresetButton, true, "(R)", Color.white, new Color(((126f * 1.5f) / 255f), ((12f * 1.5f) / 255f), ((152f * 1.5f) / 255f)), () => {
                    core.QuickLoadPluginPreset(SuperController.singleton.GetSelectedAtom(), true, true);
                }));

                if (lastButtonsCUAMCompactLayetEnabled)
                {
                    buttons.Add(new ButtonLayoutInfo(lastButtonsCUAMEnabled, false, "[CUAM]        \nReplace        ", Color.black, new Color(0.5f, 0.9f, 0.9f), () => {
                        if (CheckPersonAtomSelectedForCUAMAction())
                        {
                            core.QuickLoadReplacePreset(true);
                        }
                    }, (btn) => { btnReplace = btn; }));
                    buttons.Add(new ButtonLayoutInfo(lastButtonsCUAMEnabled, true, "(A)", Color.black, new Color(0.75f, 0.95f, 0.60f), () => {
                        if (CheckPersonAtomSelectedForCUAMAction())
                        {
                            core.QuickLoadAppendPreset(true);
                        }
                    }, (btn) => { btnAppend = btn; }));

                    buttons.Add(new ButtonLayoutInfo(lastButtonsCUAMEnabled, false, "[CUAM]        \nRem.last       ", Color.black, new Color(1.0f, 0.7f, 1.0f), () => {
                        if (CheckPersonAtomSelectedForCUAMAction())
                        {
                            core.QuickRemoveLastAddedPreset();
                        }
                    }, (btn) => { btnRemoveLast = btn; }));
                    buttons.Add(new ButtonLayoutInfo(lastButtonsCUAMEnabled, true, "(All)", Color.black, new Color(1.0f, 0.7f, 0.7f), () => {
                        if (CheckPersonAtomSelectedForCUAMAction())
                        {
                            core.QuickRemoveAllLinkedAssets();
                        }
                    }, (btn) => { btnRemoveAll = btn; }));
                }
                else
                {
                    buttons.Add(new ButtonLayoutInfo(lastButtonsCUAMEnabled, false, "[CUAM]\nReplace", Color.black, new Color(0.5f, 0.9f, 0.9f), () => {
                        if (CheckPersonAtomSelectedForCUAMAction())
                        {
                            core.QuickLoadReplacePreset(true);
                        }
                    }));
                    buttons.Add(new ButtonLayoutInfo(lastButtonsCUAMEnabled, false, "[CUAM]\nAppend", Color.black, new Color(0.75f, 0.95f, 0.60f), () => {
                        if (CheckPersonAtomSelectedForCUAMAction())
                        {
                            core.QuickLoadAppendPreset(true);
                        }
                    }));

                    buttons.Add(new ButtonLayoutInfo(lastButtonsCUAMEnabled, false, "[CUAM]\nRemove Last", Color.black, new Color(1.0f, 0.7f, 1.0f), () => {
                        if (CheckPersonAtomSelectedForCUAMAction())
                        {
                            core.QuickRemoveLastAddedPreset();
                        }
                    }));
                    buttons.Add(new ButtonLayoutInfo(lastButtonsCUAMEnabled, false, "[CUAM]\nRemove All", Color.black, new Color(1.0f, 0.7f, 0.7f), () => {
                        if (CheckPersonAtomSelectedForCUAMAction())
                        {
                            core.QuickRemoveAllLinkedAssets();
                        }
                    }));
                }

                buttons.Add(new ButtonLayoutInfo(lastButtonsCUAMToggleTriggersEnabled, false, "[CUAM]\nClothing Trg.", Color.black, new Color(0.95f, 0.75f, 0.6f), () => {
                    if (CheckPersonAtomSelectedForCUAMAction())
                    {
                        core.QuickToggleClothingTriggers();
                    }
                }));


                switch ((isDesktopMode ? lastButtonsLayout : lastButtonsVRLayout))
                {
                    case Options.QUICK_CANVAS_BUTTON_LAYOUT_STACKED:
                        {
                            float x = posx;
                            float y = posy + buttonHeight * 2;
                            int col = 0;
                            for (int i = 0; i < buttons.Count; i++)
                            {
                                ButtonLayoutInfo button = buttons[i];
                                if (button.enabled)
                                {
                                    if (button.isMergeButton)
                                    {
                                        // merge into previously button's position
                                        button.x = buttons[i - 1].x + buttonWidth * 0.63f;
                                        button.y = buttons[i - 1].y;
                                    }
                                    else
                                    {
                                        // new position
                                        button.x = x;
                                        button.y = y;

                                        col += 1;
                                        if (col >= 2)
                                        {
                                            col = 0;
                                            y -= buttonHeight;
                                            x -= buttonWidth;
                                        }
                                        else 
                                        {
                                            x += buttonWidth;
                                        }
                                    }
                                }
                            }
                        }
                        break;
                    case Options.QUICK_CANVAS_BUTTON_LAYOUT_HORIZONTAL:
                        {
                            float x = posx;
                            float y = posy;
                            for (int i = 0; i < buttons.Count; i ++)
                            {
                                ButtonLayoutInfo button = buttons[i];
                                if (button.enabled)
                                {
                                    if (button.isMergeButton)
                                    {
                                        // merge into previously button's position
                                        button.x = buttons[i - 1].x + buttonWidth * 0.63f;
                                        button.y = buttons[i - 1].y;
                                    }
                                    else
                                    {
                                        // new position
                                        button.x = x;
                                        button.y = y;
                                        x += buttonWidth;
                                    }
                                }
                            }
                        }
                        break;
                    case Options.QUICK_CANVAS_BUTTON_LAYOUT_VERTICAL:
                        {
                            float x = posx;
                            float y = posy + buttonHeight * 5;
                            for (int i = 0; i < buttons.Count; i++)
                            {
                                ButtonLayoutInfo button = buttons[i];
                                if (button.enabled)
                                {
                                    if (button.isMergeButton)
                                    {
                                        // merge into previously button's position
                                        button.x = buttons[i - 1].x + buttonWidth * 0.63f;
                                        button.y = buttons[i - 1].y;
                                    }
                                    else
                                    {
                                        // new position
                                        button.x = x;
                                        button.y = y;
                                        y -= buttonHeight;
                                    }
                                }
                            }
                        }
                        break;
                }

                
                // Create buttons
                foreach (ButtonLayoutInfo btnInfo in buttons)
                {
                    if (btnInfo.enabled)
                    {
                        UIDynamicButton btn = AddCanvasButton(btnInfo.text, () =>
                        {
                            try
                            {
                                btnInfo.callback();
                            }
                            catch (Exception e)
                            {
                                SuperController.LogError(e.ToString());
                            }
                        }, btnInfo.x, btnInfo.y, posz, btnInfo.width);
                        btn.textColor = btnInfo.textColor;
                        btn.buttonColor = btnInfo.buttonColor;
                        if (btnInfo.createdDynamicButton != null)
                        {
                            btnInfo.createdDynamicButton(btn);
                        }
                    }
                }

            });
        }

        public void CreateButtons(Action createCanvasButtons)
        {
            float worldScale = SuperController.singleton.worldScale;
            SuperController.singleton.worldScale = 1.0f;

            try
            {
                const float scale = 0.001f;

                Cleanup();
                
                isDesktopMode = !(SuperController.singleton.isOVR || SuperController.singleton.isOpenVR);

                GameObject canvasObject = new GameObject();
                canvas = canvasObject.AddComponent<Canvas>();
                canvas.renderMode = RenderMode.WorldSpace;
                canvas.pixelPerfect = false;
                SuperController.singleton.AddCanvas(canvas);

                canvasObject.layer = SuperController.singleton.mainHUD.gameObject.layer;
                canvas.transform.SetParent(SuperController.singleton.mainHUD, false);

                CanvasScaler cs = canvasObject.AddComponent<CanvasScaler>();
                cs.scaleFactor = 100.0f;
                cs.dynamicPixelsPerUnit = 1f;

                GraphicRaycaster gr = canvasObject.AddComponent<GraphicRaycaster>();

                canvas.transform.localScale = new Vector3(scale, scale, scale);
                canvas.transform.localPosition = new Vector3(0f, 0f, 0.0f);

                LookAtCamera();

                createCanvasButtons.Invoke();

                canvas.transform.Translate(0, 0, 0);
            }
            catch (Exception e)
            {
                SuperController.LogError("CreateButtons(): " + e);
                // DestroyImmediate(this);
            } 
            finally
            {
                SuperController.singleton.worldScale = worldScale;
            }
        }

        public UIDynamicButton AddCanvasButton(string name, UnityAction callback, float x, float y, float z, float width = -40, float height = 42)
        {
            // float z = (isDesktopMode ? 0f : -0.01f);
            Color accessButtonColor = new Color(1.0f, 0.4392f, 0.4392f);
            Color accessTextColor = new Color(0.0f, 0.0f, 0.0f);

            UIDynamicButton button = CreateCanvasButton(name, width, height);
            button.button.onClick.AddListener(callback);
            button.transform.Translate(x, y, z, Space.Self);
            ColorButton(button, accessTextColor, accessButtonColor);

            return button;
        }

        private UIDynamicButton CreateCanvasButton(string name, float width, float height)
        {
            Transform button = GameObject.Instantiate<Transform>(core.manager.configurableButtonPrefab);
            ConfigureTransform(button, width, height);
            ParentToCanvas(button);

            createdButtons.Add(button);

            UIDynamicButton uiButton = button.GetComponent<UIDynamicButton>();
            uiButton.label = name;
            uiButton.buttonText.fontSize = 18;
            return uiButton;
        }

        public static void ColorButton(UIDynamicButton button, Color textColor, Color buttonColor)
        {
            button.textColor = textColor;
            button.buttonColor = buttonColor;
        }

        private void ConfigureTransform(Transform t, float width, float height)
        {
            t.transform.position = Vector3.zero;
            RectTransform rt = t.GetComponent<RectTransform>();
            rt.anchoredPosition = new Vector2(width / 2, height / 2);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
        }

        private void ParentToCanvas(Transform t)
        {
            t.SetParent(canvas.transform, false);
        }

        private void LookAtCamera()
        {
            if (isDesktopMode)
            {
                canvas.transform.localEulerAngles = new Vector3(0, 180, 0);
            }
            else
            {
                if (XRSettings.enabled == false)
                {
                    Transform cameraT = SuperController.singleton.lookCamera.transform;
                    Vector3 endPos = cameraT.position + cameraT.forward * 10000000.0f;
                    canvas.transform.LookAt(endPos, cameraT.up);
                }
                else
                {
                    canvas.transform.localEulerAngles = new Vector3(28 + lastButtonsVRXRotOffset, 180 + lastButtonsVRYRotOffset, 0);
                }
            }
        }

        public void Cleanup()
        {
            try
            {
                if (createdButtons != null && createdButtons.Count > 0)
                {
                    foreach(Transform t in createdButtons)
                    {
                        if (t.gameObject != null)
                        {
                            GameObject.Destroy(t.gameObject);
                        }
                    }
                    createdButtons.Clear();
                }
                if (canvas != null)
                {
                    if (SuperController.singleton != null)
                    {
                        SuperController.singleton.RemoveCanvas(canvas);
                    }

                    canvas.transform.SetParent(null, false);

                    if (canvas.gameObject != null)
                    {
                        GameObject.Destroy(canvas.gameObject);
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Cleanup(): " + e);
                // DestroyImmediate(this);
            }
        }

        public void Destroy()
        {
            Cleanup();
        }
    }
}