using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Blazedust.CUAM
{
	public class LastAddedAssets
	{

		private Dictionary<Atom, List<Atom>> atomsPerSource;
		private Dictionary<Atom, List<DAZClothingItem>> clothingPerSource;

		public LastAddedAssets()
        {
			atomsPerSource = new Dictionary<Atom, List<Atom>>();
			clothingPerSource = new Dictionary<Atom, List<DAZClothingItem>>();

		}

		public void Clear(Atom sourceAtom)
        {
			atomsPerSource.Remove(sourceAtom);
			clothingPerSource.Remove(sourceAtom);
		}

		public bool Contains(Atom sourceAtom, Atom trackAtom)
        {
			List<Atom> atoms = null;
			return (atomsPerSource.TryGetValue(sourceAtom, out atoms) && atoms.Contains(trackAtom));
		}

		public void Remove(Atom sourceAtom, Atom trackAtom)
        {
			List<Atom> atoms = null;
			if (atomsPerSource.TryGetValue(sourceAtom, out atoms))
			{
				atoms.Remove(trackAtom);
			}
		}

		public void Add(Atom sourceAtom, Atom trackAtom)
        {
			if (sourceAtom == null)
				return;

			List<Atom> atoms = null;
			if (!atomsPerSource.TryGetValue(sourceAtom, out atoms))
			{
				atoms = new List<Atom>();
				atomsPerSource.Add(sourceAtom, atoms);
			}
			atoms.Add(trackAtom);
		}

		public List<Atom> Get(Atom sourceAtom)
        {
			List<Atom> atoms = null;
			if (!atomsPerSource.TryGetValue(sourceAtom, out atoms))
			{
				return new List<Atom>();
			}
			return atoms;
		}

		public void AddClothing(Atom sourceAtom, DAZClothingItem clothingItem)
        {
			List<DAZClothingItem> clothings = null;
			if (!clothingPerSource.TryGetValue(sourceAtom, out clothings))
            {
				clothings = new List<DAZClothingItem>();
				clothingPerSource.Add(sourceAtom, clothings);
			}
			clothings.Add(clothingItem);
		}

		public List<DAZClothingItem> GetClothing(Atom sourceAtom)
        {
			List<DAZClothingItem> clothings = null;
			if (!clothingPerSource.TryGetValue(sourceAtom, out clothings))
			{
				return new List<DAZClothingItem>();
			}
			return clothings;
		}
		
	}
}