using System;
using System.Text;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Blazedust.CUAM
{
	public class Notifier
	{

		private static List<KeyValuePair<string, string>> targetPluginActions = null;

		/// <summary>
        /// Checks if the target atom has any supported plugins with actions registered to run after atom update.
        /// </summary>
		public static void NotifyAtomUpdatedToPlugins(Atom receivingAtom)
		{
			if (targetPluginActions == null)
            {
				targetPluginActions = new List<KeyValuePair<string, string>>();
				// KeyValuePair of Plugin -> Action (Action in plugin registered with RegisterAction(...) with a JSONStorableAction with the listed ID)
				targetPluginActions.Add(new KeyValuePair<string, string>("Blazedust.ParentHoldLink", "resetPosRot"));
			}
			foreach (string receiverID in receivingAtom.GetStorableIDs())
			{
				if (receiverID.StartsWith("plugin#"))
                {
					foreach (KeyValuePair<string, string> pluginAction in targetPluginActions.Where((x) => { return receiverID.EndsWith(x.Key); })) 
					{
						// Potential plugin!
						JSONStorable receiver = receivingAtom.GetStorableByID(receiverID);
						if (receiver != null)
						{
							foreach (string actionParam in receiver.GetActionNames())
							{
								if (pluginAction.Value == actionParam)
                                {
									receiver.CallAction(actionParam);
									break;
								}
							}
						}
					}
				}
			}
		}

		public static void NotifyAtomColorParameter(Atom receivingAtom, string pluginId, string colorParam, HSVColor c)
		{
			foreach (string receiverID in receivingAtom.GetStorableIDs())
			{
				if (receiverID == pluginId)
				{
					JSONStorable receiver = receivingAtom.GetStorableByID(receiverID);
					if (receiver != null)
					{
						receiver.SetColorParamValue(colorParam, c);
						break;
					}
				}
			}
		}

	}
}