using System;
using System.Text;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Blazedust.CUAM
{
	public class Options
	{
		CustomUnityAssetManager core;

		public JSONStorableBool autoSelectPersonFromScene;
		public JSONStorableBool autoSelectAssetController;
		public JSONStorableBool delayPhysicsProperties;
		public JSONStorableBool enableScalingModifier;
		public JSONStorableBool preservePersonScaleOnChange;
		public JSONStorableBool ignoreAllSceneAssets;
		public JSONStorableBool ignoreSceneAssetsContainingScripts;
		public JSONStorableBool ignoreSceneAssetsMarkedHidden;
		public JSONStorableBool smartRemovalOfPresets;
		public JSONStorableBool enableLoadTriggers;
		public JSONStorableBool updateScriptAtomUids;
		public JSONStorableBool enablePrioritizedLoadingTriggers;
		public JSONStorableBool enableCUAReuseOnLoad;
		public JSONStorableBool enablePlayModeUpdates;
		public JSONStorableBool verbosePluginLogs;
		public JSONStorableBool showLegacyPersonUidTriggerSelector;
		public JSONStorableBool ignoreDisabledLinkedAtoms;
		public JSONStorableString autoLoaderPresetsList;
		public JSONStorableStringChooser autoLoaderPresetsListPageSelector;
		public JSONStorableString includedAtomTypesList;
		public JSONStorableBool quickCanvasVAMPresetButtons;
		public JSONStorableBool quickCanvasVAMPluginPresetButton;
		public JSONStorableBool quickCanvasVAMMergePresetsButtons;
		public JSONStorableBool quickCanvasCUAMButtons;
		public JSONStorableBool quickCanvasCUAMButtonsCompactLayout;
		public JSONStorableBool quickCanvasCUAMToggleTriggers;
		public JSONStorableStringChooser quickCanvasButtonsType;
		public JSONStorableFloat quickCanvasButtonsXOffset;
		public JSONStorableFloat quickCanvasButtonsYOffset;

		public JSONStorableStringChooser quickCanvasButtonsVRType;
		public JSONStorableFloat quickCanvasButtonsVRXOffset;
		public JSONStorableFloat quickCanvasButtonsVRYOffset;
		public JSONStorableFloat quickCanvasButtonsVRZOffset;
		public JSONStorableFloat quickCanvasButtonsVRXRotOffset;
		public JSONStorableFloat quickCanvasButtonsVRYRotOffset;

		public JSONStorableString addLoadTriggerPreviewInfo;
		public JSONStorableString addLoadTriggerPluginPresetValue;
		public JSONStorableString addLoadTriggerCUAMPresetValue;
		public JSONStorableStringChooser addLoadTriggerPersonUidSelection;
		public JSONStorableStringChooser addLoadTriggerMorphSelection;
		public JSONStorableStringChooser addLoadTriggerSorting;
		public JSONStorableStringChooser removeLoadTriggersSelection;
		private JSONStorableStringChooser includedAtomTypeSelection;

		private List<AutoLoadTriggerInfo> autoLoadTriggerPresets;
		private List<AutoLoadTriggerListInfo> autoLoadTriggerPresetsFiltered;
		private HashSet<string> autoLoadTriggerMorphNames;
		private bool resetActiveMorphsAfterLoading;
		private int autoLoadCheckerCounter;
		private Dictionary<Atom, AtomTriggerStates> activeTriggersPerAtom;
		private List<string> includedAtomTypes;
		private DateTime optionsDebounceSaveTime;
		private bool optionsDebounceSaveEnabled;

		public const string ANY_MORPH = "* ANY *";

		public const string QUICK_CANVAS_BUTTON_LAYOUT_STACKED = "Stacked";
		public const string QUICK_CANVAS_BUTTON_LAYOUT_HORIZONTAL = "Horizontal";
		public const string QUICK_CANVAS_BUTTON_LAYOUT_VERTICAL = "Vertical";		

		private Atom lastRootAtom;
		private UIHelper uiHelper;

		public Options(CustomUnityAssetManager c)
        {
			core = c;
			uiHelper = new UIHelper(c);

			resetActiveMorphsAfterLoading = true;
			autoLoadTriggerPresets = new List<AutoLoadTriggerInfo>();
			autoLoadTriggerPresetsFiltered = new List<AutoLoadTriggerListInfo>();
			autoLoadTriggerMorphNames = new HashSet<string>();
			activeTriggersPerAtom = new Dictionary<Atom, AtomTriggerStates>();
			autoLoadCheckerCounter = 0;
			lastRootAtom = null;
			includedAtomTypes = new List<string>();

			autoSelectPersonFromScene = new JSONStorableBool("cuabAutoSelectPerson", true, SelectedOptionsToggleChanged);
			autoSelectAssetController = new JSONStorableBool("cuabOptionAutoSelectAssetController", true, SelectedOptionsToggleChanged);
			delayPhysicsProperties = new JSONStorableBool("cuabOptionDelayPhysicsProperties", true, SelectedOptionsToggleChanged);
			enableScalingModifier = new JSONStorableBool("cuabOptionEnableScaling", true, SelectedOptionsToggleChanged);
			preservePersonScaleOnChange = new JSONStorableBool("cuabOptionPreservePersonScaleOnChange", true, SelectedOptionsToggleChanged);
			enableLoadTriggers = new JSONStorableBool("cuabOptionEnableLoadTriggers", true, (bool val) => {
				SelectedOptionsToggleChanged(val);
				UpdateAutoLoaderPresetsList();
			});
			ignoreAllSceneAssets = new JSONStorableBool("cuabOptionIgnoreAllSceneAssets", true, SelectedOptionsToggleChanged);
			ignoreSceneAssetsContainingScripts = new JSONStorableBool("cuabOptionIgnoreSceneAssetsContainingScripts", true, SelectedOptionsToggleChanged);
			ignoreSceneAssetsMarkedHidden = new JSONStorableBool("cuabOptionIgnoreSceneAssetsMarkedHidden", true, SelectedOptionsToggleChanged);
			smartRemovalOfPresets = new JSONStorableBool("cuabOptionSmartRemovalOfPresets", true, SelectedOptionsToggleChanged);
			verbosePluginLogs = new JSONStorableBool("cuabVerbosePluginLogs", false, SelectedOptionsToggleChanged);
			showLegacyPersonUidTriggerSelector = new JSONStorableBool("cuabShowLegacyPersonUidTriggerSelector", false, SelectedOptionsToggleChanged);
			autoLoaderPresetsList = new JSONStorableString("cuabAutoLoaderPresetsList", "");
			includedAtomTypesList = new JSONStorableString("cuabIncludedAtomTypesList", "");
			updateScriptAtomUids = new JSONStorableBool("cuabUpdateScriptAtomUids", true, SelectedOptionsToggleChanged);
			enablePrioritizedLoadingTriggers = new JSONStorableBool("cuabEnablePrioritizedLoadingTriggers", false, SelectedOptionsToggleChanged);
			enableCUAReuseOnLoad = new JSONStorableBool("enableCUAReuseOnLoad", true, SelectedOptionsToggleChanged);
			enablePlayModeUpdates = new JSONStorableBool("cuabEnablePlayModeUpdates", false, SelectedOptionsToggleChanged);
			ignoreDisabledLinkedAtoms = new JSONStorableBool("cuabIgnoreDisabledLinkedAtoms", true, SelectedOptionsToggleChanged);

			quickCanvasVAMPresetButtons = new JSONStorableBool("cuabQuickCanvasVAMPresetsButtons", true, SelectedOptionsToggleChanged);
			quickCanvasVAMPluginPresetButton = new JSONStorableBool("cuabQuickCanvasVAMPluginPresetsButtons", true, SelectedOptionsToggleChanged);
			quickCanvasVAMMergePresetsButtons = new JSONStorableBool("cuabQuickCanvasVAMMergePresetsButton", true, SelectedOptionsToggleChanged);
			quickCanvasCUAMButtons = new JSONStorableBool("cuabQuickCanvasCUAMButtons", true, SelectedOptionsToggleChanged);
			quickCanvasCUAMButtonsCompactLayout = new JSONStorableBool("cuabQuickCanvasCUAMButtonsCompactLayout", true, SelectedOptionsToggleChanged);
			quickCanvasCUAMToggleTriggers = new JSONStorableBool("cuabQuickCanvasCUAMToggleTriggers", false, SelectedOptionsToggleChanged);

			List<string> quickCanvasButtonLayouts = new List<string>();
			quickCanvasButtonLayouts.Add(QUICK_CANVAS_BUTTON_LAYOUT_STACKED);
			quickCanvasButtonLayouts.Add(QUICK_CANVAS_BUTTON_LAYOUT_HORIZONTAL);
			quickCanvasButtonLayouts.Add(QUICK_CANVAS_BUTTON_LAYOUT_VERTICAL);
			quickCanvasButtonsType = new JSONStorableStringChooser("cuabQuickCanvasButtonsType", quickCanvasButtonLayouts, QUICK_CANVAS_BUTTON_LAYOUT_STACKED, "Desktop Button Layout", (string val) => {
				SaveOptions();
			});
			quickCanvasButtonsXOffset = new JSONStorableFloat("cuabQuickCanvasPosXOffset", 0.0f, (JSONStorableFloat jf) => {
				SaveOptions(1f); // debounce save 1 second if case of many changes
			}, -4f, 4f, false);
			quickCanvasButtonsYOffset = new JSONStorableFloat("cuabQuickCanvasPosYOffset", 0.0f, (JSONStorableFloat jf) => {
				SaveOptions(1f); // debounce save 1 second if case of many changes
			}, -4f, 4f, false);

			quickCanvasButtonsVRType = new JSONStorableStringChooser("cuabQuickCanvasButtonsVRType", quickCanvasButtonLayouts, QUICK_CANVAS_BUTTON_LAYOUT_HORIZONTAL, "VR Button Layout", (string val) => {
				SaveOptions();
			});
			quickCanvasButtonsVRXOffset = new JSONStorableFloat("cuabQuickCanvasButtonsVRXOffset", 0.236f, (JSONStorableFloat jf) => {
				SaveOptions(1f); // debounce save 1 second if case of many changes
			}, -4f, 4f, false);
			quickCanvasButtonsVRYOffset = new JSONStorableFloat("cuabQuickCanvasButtonsVRYOffset", -0.106f, (JSONStorableFloat jf) => {
				SaveOptions(1f); // debounce save 1 second if case of many changes
			}, -4f, 4f, false);
			quickCanvasButtonsVRZOffset = new JSONStorableFloat("cuabQuickCanvasButtonsVRZOffset", 0.0f, (JSONStorableFloat jf) => {
				SaveOptions(1f); // debounce save 1 second if case of many changes
			}, -4f, 4f, false);

			quickCanvasButtonsVRXRotOffset = new JSONStorableFloat("cuabQuickCanvasButtonsVRXRotOffset", 0.0f, (JSONStorableFloat jf) => {
				SaveOptions(1f); // debounce save 1 second if case of many changes
			}, -180f, 180f, false);
			quickCanvasButtonsVRYRotOffset = new JSONStorableFloat("cuabQuickCanvasButtonsVRYRotOffset", 0.0f, (JSONStorableFloat jf) => {
				SaveOptions(1f); // debounce save 1 second if case of many changes
			}, -180f, 180f, false);

			optionsDebounceSaveEnabled = false;
			optionsDebounceSaveTime = DateTime.MinValue;
		}

		public void Destroy()
		{
			uiHelper.Destroy();
			uiHelper = null;
			lastRootAtom = null;
			core = null;
			activeTriggersPerAtom.Clear();
			activeTriggersPerAtom = null;
			autoLoadTriggerPresets.Clear();
			autoLoadTriggerPresets = null;
			autoLoadTriggerPresetsFiltered.Clear();
			autoLoadTriggerPresetsFiltered = null;
			autoLoadTriggerMorphNames.Clear();
			autoLoadTriggerMorphNames = null;
		}

		public void InitStorables()
		{
			addLoadTriggerMorphSelection = new JSONStorableStringChooser("cuabAddTriggerMorphSelection", GetMorphTriggerNames(), "", "Trigger Morph", (string triggerMorphName) => { UpdateAddTriggerPreviewInfoText(); });
			
			addLoadTriggerPersonUidSelection = new JSONStorableStringChooser("cuabAddTriggerPersonUidSelection", GetPersonUidTriggerNames(), "", "(Optional) Trigger Person Uid", (string personUid) => { UpdateAddTriggerPreviewInfoText(); });

			addLoadTriggerPluginPresetValue = new JSONStorableString("cuabAddLoadTriggerPluginPresetValue", "");
			addLoadTriggerCUAMPresetValue = new JSONStorableString("cuabAddLoadTriggerCUAMPresetValue", "");
			addLoadTriggerPreviewInfo = new JSONStorableString("cuabAddLoadTriggerPreviewInfo", "");

			removeLoadTriggersSelection = new JSONStorableStringChooser("cuabRemoveTriggerSelection", GetLoadTriggerNames(true), "", "Selected Trigger");

			addLoadTriggerSorting = new JSONStorableStringChooser("addLoadTriggerSorting", GetTriggerSortingOptions(), TRIGGER_SORT_TRIGGER_DEFAULT, "Sort/Filter Triggers", (string triggerSorting) => {
				autoLoaderPresetsListPageSelector.choices = GetTriggerPageOptions();
				autoLoaderPresetsListPageSelector.valNoCallback = autoLoaderPresetsListPageSelector.choices[0];
				UpdateAutoLoaderPresetsList();
			});

			{
				List<string> pageOptions = GetTriggerPageOptions();
				autoLoaderPresetsListPageSelector = new JSONStorableStringChooser("cuabAutoLoaderPresetsListPageSelector", pageOptions, pageOptions[0], "Triggers Page", (string val) => { AutoLoaderPresetsListPageSelectorUpdated(); });
			}

			List<string> availableAtomTypes = new List<string>();
			// availableAtomTypes.Add("Animation/AnimationPattern"); // Status: WIP/RESEARCH
			// availableAtomTypes.Add("CustomUnityAsset"); <-- Forced included, always
			availableAtomTypes.Add("Light/InvisibleLight");
			availableAtomTypes.Add("Light/Torch");

			availableAtomTypes.Add("Misc/Empty");
			availableAtomTypes.Add("Misc/ImagePanel");
			availableAtomTypes.Add("Misc/ImagePanelEmissive");
			availableAtomTypes.Add("Misc/SimpleSign");
			availableAtomTypes.Add("Misc/SubScene"); // Need to call Receiver: SubScene with action: LoaddSubScene after it's setup!
			availableAtomTypes.Add("Misc/UIText");
			// Browser might not make too much sense for the purpose of this plugin but add anyway...
			availableAtomTypes.Add("Misc/WebBrowser");
			availableAtomTypes.Add("Misc/WebPanel");
			availableAtomTypes.Add("Misc/WebPanelEmissive");

			availableAtomTypes.Add("Props/AptBook01");
			availableAtomTypes.Add("Props/AptBook02");
			availableAtomTypes.Add("Props/SkullQueenSword");

			availableAtomTypes.Add("Toys/Dildo");
			availableAtomTypes.Add("Toys/Paddle");
			availableAtomTypes.Add("Toys/ToyAH");
			availableAtomTypes.Add("Toys/ToyBP");

			availableAtomTypes.Add("Shapes/Capsule");
			availableAtomTypes.Add("Shapes/Cube");
			availableAtomTypes.Add("Shapes/ISCapsule");
			availableAtomTypes.Add("Shapes/ISCone");
			availableAtomTypes.Add("Shapes/ISCube");
			availableAtomTypes.Add("Shapes/ISCylinder");
			availableAtomTypes.Add("Shapes/ISSphere");
			availableAtomTypes.Add("Shapes/ISTube");
			availableAtomTypes.Add("Shapes/Sphere");

			// Browser might not make too much sense for the purpose of this plugin but add anyway...
			availableAtomTypes.Add("Triggers/Button");
			availableAtomTypes.Add("Triggers/CollisionTrigger");
			availableAtomTypes.Add("Triggers/LookAtTrigger");
			availableAtomTypes.Add("Triggers/UIButton");
			availableAtomTypes.Add("Triggers/UISlider");
			availableAtomTypes.Add("Triggers/UIToggle");
			availableAtomTypes.Add("Triggers/VariableTrigger");

			availableAtomTypes.Add("Force/CycleForce"); // cycle force for breathing kind of presets.
			// availableAtomTypes.Add("Force/GrabPoint"); // <-- Can't get this one to work properly. Never add to this plugin! F**ks up everything! :)
			availableAtomTypes.Add("Force/SyncForce");
			
			// availableAtomTypes.Add("People/Person"); // <-- Not supported at all, plugin uses Person as a root atom. Never add!
			includedAtomTypeSelection = new JSONStorableStringChooser("cuabAddAtomTypeSelection", availableAtomTypes, "", "Select Atom Type");
			
		}

		public void CreateGUIAutoLoadTriggers(UIHelper uiHelper)
        {
			UIDynamicToggle toggle;
			UIDynamicButton btn;
			UIDynamicTextField text;
			UIDynamicPopup popup;

			btn = uiHelper.CreateButton("Manage auto load triggers: ", false);
			btn.buttonColor = Color.yellow;
			btn.button.onClick.AddListener(() =>
			{
				ManageAutoLoadTriggersHelp();
			});

			// Trigger Morph
			popup = uiHelper.CreateFilterablePopup(addLoadTriggerMorphSelection, false);
			popup.popupPanelHeight = 700f;
			addLoadTriggerMorphSelection.popup.onOpenPopupHandlers += () => {
				addLoadTriggerMorphSelection.choices = GetMorphTriggerNames();
			};

			// (Optional) Trigger Person Uid
			if (showLegacyPersonUidTriggerSelector.val)
			{
				popup = uiHelper.CreateScrollablePopup(addLoadTriggerPersonUidSelection, false);
				popup.popupPanelHeight = 700f;
				addLoadTriggerPersonUidSelection.popup.onOpenPopupHandlers += () =>
				{
					addLoadTriggerPersonUidSelection.choices = GetPersonUidTriggerNames();
				};
			}

			btn = uiHelper.CreateButton("Select CUAM Preset", false);
			btn.buttonColor = new Color(0.65f, 0.75f, 0.95f);
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					SelectTriggerLoadCUAMPresetUsingFileBrowser();
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			btn = uiHelper.CreateButton("Select Plugin Preset", false);
			btn.buttonColor = new Color(0.65f, 0.75f, 0.95f);
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					// Using FileBrowser to select a preset. Easier this way!
					SelectTriggerLoadPluginPresetUsingFileBrowser();
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			// Info field for selected preset to load
			UIDynamicTextField textfield = uiHelper.CreateTextField(addLoadTriggerPreviewInfo, false);
			UIHelper.SetTextFieldHeight(textfield, 200);


			btn = uiHelper.CreateButton("Add/Create Trigger", false);
			btn.buttonColor = new Color(0.7f, 1.0f, 0.7f);
			btn.button.onClick.AddListener(() =>
			{
				try
				{
                    if (addLoadTriggerMorphSelection.val == null || addLoadTriggerMorphSelection.val == "")
                    {
                        SuperController.LogMessage("Select a Trigger Morph!");
                    }
                    else if ((addLoadTriggerCUAMPresetValue.val == null || addLoadTriggerCUAMPresetValue.val == "") && (addLoadTriggerPluginPresetValue.val == null || addLoadTriggerPluginPresetValue.val == ""))
                    {
                        SuperController.LogMessage("Select a Preset!");
                    }
                    else
                    {
                        string presetName = "";
                        string presetType = PRESET_CUAM;
                        if (addLoadTriggerCUAMPresetValue.val != "")
                        {
                            presetName = addLoadTriggerCUAMPresetValue.val;
							presetType = PRESET_CUAM;
						}
                        else if (addLoadTriggerPluginPresetValue.val != "")
                        {
                            presetName = addLoadTriggerPluginPresetValue.val;
							presetType = PRESET_PLUGIN;

							// Clear existing plugins matching the plugin preset and let the trigger the plugin preset. 
							// This is done to avoid having duplicate plugins as we expect the user having the plugin preset they want to add a trigger for on the look already!
							core.pluginHelper.ClearPluginsFromPreset(core.rootAtom, presetName, false);
						}

                        AutoLoadTriggerInfo newLoadTrigger = AddLoadTrigger(addLoadTriggerPersonUidSelection.val, addLoadTriggerMorphSelection.val, presetName, presetType);
                        addLoadTriggerPersonUidSelection.val = "";
                        addLoadTriggerMorphSelection.valNoCallback = "";
                        addLoadTriggerCUAMPresetValue.valNoCallback = "";
                        addLoadTriggerPluginPresetValue.valNoCallback = "";
						UpdateAddTriggerPreviewInfoText();

                        core.DoNextFrame(() =>
                        {
                            SetSelectedAutoLoadTriggerInfo(newLoadTrigger);
                        }, 30);
                    }

					autoLoaderPresetsListPageSelector.choices = GetTriggerPageOptions();
					autoLoaderPresetsListPageSelector.val = autoLoaderPresetsListPageSelector.choices[0];
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			if (showLegacyPersonUidTriggerSelector.val)
			{
				uiHelper.CreateSpacer(false, 80);
			} 
			else
            {
				uiHelper.CreateSpacer(false, 190);
			}

			// Selected Trigger
			removeLoadTriggersSelection = new JSONStorableStringChooser("cuabRemoveTriggerSelection", GetLoadTriggerNames(true), "", "Selected Trigger");
			popup = uiHelper.CreateFilterablePopup(removeLoadTriggersSelection, false);
			popup.popupPanelHeight = 740f;
			removeLoadTriggersSelection.popup.onOpenPopupHandlers += () => {
				RefreshRemoveLoadSelectedTriggersList();
			};
			uiHelper.OpenUp(popup);

			btn = uiHelper.CreateButton("Test Trigger\n(Load preset from selected trigger)", false);
			btn.buttonColor = new Color(0.75f, 0.65f, 0.95f);
			btn.height = 80;
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					// Append is safer to do as it adds atoms (not remove anything). Choke it a bit though.
					if (core.loading.LoadingInProgress(90))
					{
						return;
					}
					if (!core.LogicEnabled)
					{
						SuperController.LogMessage(CustomUnityAssetManager.PLUGIN_NOT_ENABLED);
						return;
					}
					AutoLoadTriggerInfo autoLoadTrigger = GetSelectedAutoLoadTriggerInfo();
					if (autoLoadTrigger != null)
					{	
						if (autoLoadTrigger.presetType == PRESET_CUAM)
						{
							string file = core.pluginPath.GetFileFromPresetSelection(autoLoadTrigger.presetName);
							if (file != "")
							{
								SuperController.LogMessage("CUAM: Testing loading trigger CUAM preset: " + autoLoadTrigger.presetName);
								core.LoadPresetFromFile(file, core.rootAtom, false);
							}
                        }
                        else if (autoLoadTrigger.presetType == PRESET_PLUGIN)
                        {
							SuperController.LogMessage("CUAM: Testing loading trigger plugin preset: " + autoLoadTrigger.presetName);
							core.pluginHelper.QueueLoadPluginPreset(core.rootAtom, autoLoadTrigger);
						}
					}
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			btn = uiHelper.CreateButton("Delete Selected Trigger", false);
			btn.buttonColor = new Color(1.0f, 0.5f, 0.5f);
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					if (removeLoadTriggersSelection.val != "")
					{
						RemoveLoadTriggerName(removeLoadTriggersSelection.val);
						removeLoadTriggersSelection.valNoCallback = "";

						RefreshRemoveLoadSelectedTriggersList();
					}
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			// Sort/Filter
			popup = uiHelper.CreatePopup(addLoadTriggerSorting, true);
			popup.popupPanelHeight = 400f;
			addLoadTriggerSorting.popup.onOpenPopupHandlers += () => {
				addLoadTriggerSorting.choices = GetTriggerSortingOptions();
			};

			// Sort list page selector
			popup = uiHelper.CreatePopup(autoLoaderPresetsListPageSelector, true);
			popup.popupPanelHeight = 400f;
			autoLoaderPresetsListPageSelector.popup.onOpenPopupHandlers += () => {
				autoLoaderPresetsListPageSelector.choices = GetTriggerPageOptions();
			};
			autoLoaderPresetsListPageSelector.choices = GetTriggerPageOptions();
			autoLoaderPresetsListPageSelector.val = autoLoaderPresetsListPageSelector.choices[0];

			// Sort list
			text = uiHelper.CreateTextField(autoLoaderPresetsList, true);
			text.height = 900;
			

			uiHelper.CreateSpacer(false, 30);

			UpdateAddTriggerPreviewInfoText();
		}

		public void CreateGUIOptions(UIHelper uiHelper)
		{
			UIDynamicToggle toggle;
			UIDynamicButton btn;
			UIDynamicTextField text;
			UIDynamicPopup popup;

			// Options section
			// uiHelper.CreateSpacer(true, 50);
			btn = uiHelper.CreateButton("Plugin options: ", false);
			btn.buttonColor = Color.yellow;
			btn.button.onClick.AddListener(() =>
			{
				OptionsHelp();
			});

			toggle = uiHelper.CreateToggle(autoSelectPersonFromScene, false);
			toggle.labelText.text = "Auto Select Person From Scene";

			toggle = uiHelper.CreateToggle(autoSelectAssetController, false);
			toggle.label = "Auto Select Asset Controller";

			toggle = uiHelper.CreateToggle(delayPhysicsProperties, false);
			toggle.height = 80;
			toggle.label = "Delay Physic Properties On Load Preset";

			toggle = uiHelper.CreateToggle(enableScalingModifier, false);
			toggle.height = 80;
			toggle.label = "Enable Scaling Modifier For Assets";

			toggle = uiHelper.CreateToggle(preservePersonScaleOnChange, false);
			toggle.height = 120;
			toggle.label = "Preserve Person Scale On Appearance Change\n(Quick Canvas Button Only!)";

			toggle = uiHelper.CreateToggle(updateScriptAtomUids, false);
			toggle.height = 80;
			toggle.label = "Update AtomUids In Scripts And Properties";

			toggle = uiHelper.CreateToggle(enablePrioritizedLoadingTriggers, false);
			toggle.height = 40;
			toggle.label = "Enable Prioritized Loading Triggers";

			toggle = uiHelper.CreateToggle(enableCUAReuseOnLoad, false);
			toggle.height = 80;
			toggle.label = "Reuse CUA AssetUrl On Load\nIf Same Filename";
			// Reuse CUA AssetUrl On Load If Same Filename

			toggle = uiHelper.CreateToggle(enablePlayModeUpdates, false);
			toggle.height = 40;
			toggle.label = "Enable Update In Play Mode";

			toggle = uiHelper.CreateToggle(ignoreDisabledLinkedAtoms, false);
			toggle.height = 40;
			toggle.label = "Ignore Disabled Linked Atoms When Saving Presets";

			toggle = uiHelper.CreateToggle(ignoreAllSceneAssets, true);
			toggle.height = 80;
			toggle.label = "Filter for Replace/Remove:\nIgnore All Scene Assets";
			toggle.backgroundColor = new Color(0.9f, 0.8f, 0.8f);

			toggle = uiHelper.CreateToggle(ignoreSceneAssetsContainingScripts, true);
			toggle.height = 120;
			toggle.label = "Filter for Replace/Remove:\nIgnore Scene Assets Containing Scripts";
			toggle.backgroundColor = new Color(0.9f, 0.8f, 0.8f);

			toggle = uiHelper.CreateToggle(ignoreSceneAssetsMarkedHidden, true);
			toggle.height = 120;
			toggle.label = "Filter for Replace/Remove:\nIgnore Scene Assets Marked Hidden";
			toggle.backgroundColor = new Color(0.9f, 0.8f, 0.8f);

			uiHelper.CreateSpacer(false, 40);
			uiHelper.CreateSpacer(true, 110);

			toggle = uiHelper.CreateToggle(verbosePluginLogs, true);
			toggle.height = 80;
			toggle.label = "Verbose plugin events and statuses";
			toggle.backgroundColor = new Color(0.6f, 0.9f, 0.9f);

			toggle = uiHelper.CreateToggle(showLegacyPersonUidTriggerSelector, true);
			toggle.height = 80;
			toggle.label = "Show Deprecated 'Trigger Person Uid' in Auto Load Triggers";
			toggle.backgroundColor = new Color(0.7f, 0.5f, 0.5f);

			toggle = uiHelper.CreateToggle(smartRemovalOfPresets, true);
			toggle.height = 80;
			toggle.label = "Enable Smart Auto-Load-Trigger Preset Removal";
			toggle.backgroundColor = new Color(0.8f, 0.9f, 0.9f);

			toggle = uiHelper.CreateToggle(enableLoadTriggers, true);
			toggle.label = "Enable Load Triggers";
			toggle.backgroundColor = new Color(0.8f, 0.9f, 0.9f);

			uiHelper.CreateSpacer(true, 20);
			uiHelper.CreateSpacer(false, 20);


			btn = uiHelper.CreateButton("Atom types to handle (assets):", false);
			btn.buttonColor = Color.yellow;
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					SuperController.singleton.ClearMessages();
					SuperController.LogMessage("* Atom types to handle");
					SuperController.LogMessage("");
					SuperController.LogMessage("Add or Remove any additional atom types this plugin should handle.");
					SuperController.LogMessage("");
					SuperController.LogMessage("By default only CustomUnityAssets are handled but you can add other atom types here.");
					SuperController.LogMessage("");
					SuperController.singleton.OpenMessageLogPanel();
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			// Select Atom Type
			popup = uiHelper.CreateFilterablePopup(includedAtomTypeSelection, false);
			popup.popupPanelHeight = 600f;

			btn = uiHelper.CreateButton("Add Atom Type", false);
			btn.buttonColor = new Color(0.7f, 1.0f, 0.7f);
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					string atomType = includedAtomTypeSelection.val;
					if (atomType != null && atomType != "" && atomType.Contains("/"))
						atomType = atomType.Substring(atomType.IndexOf("/") + 1);

					if (atomType != null && atomType != "" && !includedAtomTypes.Contains(atomType))
					{
						includedAtomTypes.Add(atomType);
						SelectedOptionsToggleChanged(true); // trigger save
						UpdateIncludedAtomTypesList();
					}
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			btn = uiHelper.CreateButton("Remove Atom Type", false);
			btn.buttonColor = new Color(1.0f, 0.7f, 0.7f);
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					string atomType = includedAtomTypeSelection.val;
					if (atomType != null && atomType != "" && atomType.Contains("/"))
						atomType = atomType.Substring(atomType.IndexOf("/") + 1);

					if (atomType != null && atomType != "" && includedAtomTypes.Contains(atomType))
					{
						includedAtomTypes.Remove(atomType);
						SelectedOptionsToggleChanged(true); // trigger save
						UpdateIncludedAtomTypesList();
					}
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			text = uiHelper.CreateTextField(includedAtomTypesList, false);
			text.height = 450;

			uiHelper.CreateSpacer(true, 760);

			btn = uiHelper.CreateButton("Plugin quirks (?)", true);
			btn.buttonColor = Color.yellow;
			btn.button.onClick.AddListener(() =>
			{
				PluginQuirksHelp();
			});

			//btn = uiHelper.CreateButton("External option files (?)", true);
			//btn.buttonColor = Color.yellow;
			//btn.button.onClick.AddListener(() =>
			//{
			//	try
			//	{
			//		SuperController.singleton.ClearMessages();
			//		SuperController.LogMessage("For creators: This plugin reads all option_* files and preset_* files in the plugin's preset folder. This allows you to share options and preset files for this plugin.");
			//		SuperController.LogMessage("");
			//		SuperController.LogMessage("All your preset files starting with 'preset_' can be shared directly.");
			//		SuperController.LogMessage("");
			//		SuperController.LogMessage("Option files are shared by creating a file named 'options_MYOPTIONS.json' in this plugin's preset folder.");
			//		SuperController.LogMessage("");
			//		SuperController.LogMessage("A valid shared options file look like this:");
			//		SuperController.LogMessage("");
			//		SuperController.LogMessage("{");
			//		SuperController.LogMessage("  \"autoMorphLoaders\" : [ ");
			//		SuperController.LogMessage("     {");
			//		SuperController.LogMessage("        \"id\": \"MorphName\",");
			//		SuperController.LogMessage("        \"personUid\": \"\",");
			//		SuperController.LogMessage("        \"preset\": \"NameOfPreset\"");
			//		SuperController.LogMessage("     }");
			//		SuperController.LogMessage("  ],");
			//		SuperController.LogMessage("  \"includedAtomTypes\": \"CustomUnityAsset,ImagePanel\"");
			//		SuperController.LogMessage("}");
			//		SuperController.LogMessage("");
			//		SuperController.LogMessage("Where NameOfPreset matches \"preset_NameOfPreset.json\" file and id matches the morph name. personUid is optional and you should leave it with an empty string.");
			//		SuperController.LogMessage("Simply copy your own options.json file after configuring an auto load trigger and only keep the relevant autoMorphLoaders entries as seen in the above example. No other configuration keys will be read in from shared options files.");
			//		SuperController.LogMessage("");
			//		SuperController.LogMessage("Note: The plugin files must be in this plugin's preset folder, so it won't support .vac or .var packaging. You can always share the plugin files in a separate .zip file.");
			//		SuperController.singleton.OpenMessageLogPanel();
			//	}
			//	catch (Exception e)
			//	{
			//		SuperController.LogError(e.ToString());
			//	}
			//});
		}

		public void CreateGUIQuickCanvasButtons(UIHelper uiHelper)
		{
			UIDynamicToggle toggle;
			UIDynamicButton btn;
			UIDynamicTextField text;
			UIDynamicPopup popup;

			btn = uiHelper.CreateButton("Quick canvas buttons:", false);
			btn.buttonColor = Color.yellow;
			btn.button.onClick.AddListener(() =>
			{
				try
				{
					SuperController.singleton.ClearMessages();
					SuperController.LogMessage("* Quick canvas buttons");
					SuperController.LogMessage("");
					SuperController.LogMessage("Quick canvas buttons for loading/removing CustomUnityAssetManager presets in edit mode for the selected person in the scene without having to navigate to the session plugin interface.");
					SuperController.LogMessage("");
					SuperController.LogMessage("It includes quick canvas buttons for VAM's built in Appearance Preset, Clothing Preset and Plugin Preset to avoid having to click around in the interface too much when changing appearance/clothing and updating plugins (which I do a lot).");
					SuperController.LogMessage("");
					SuperController.LogMessage("Here you can configure where to show the buttons and how. The buttons only do something when you have a person atom selected.");
					SuperController.LogMessage("");
					SuperController.LogMessage("Select Clothing: Select VAM clothing preset.");
					SuperController.LogMessage("Select Clothing (M): Select VAM clothing preset using merge load.");
					SuperController.LogMessage("Select Appearance: Select VAM appearance preset.");
					SuperController.LogMessage("Merge Plugin: Select VAM plugin preset using merge load.");
					SuperController.LogMessage("Merge Plugin (R): Select VAM plugin preset using merge load but also clearing any plugins from the preset, effectively replacing existing ones. Useful if you want to apply your own preset of Glance, Embody, ImprovedPov or other plugins intended to bring more life into characters.");
					SuperController.LogMessage("");
					SuperController.singleton.OpenMessageLogPanel();
				}
				catch (Exception e)
				{
					SuperController.LogError(e.ToString());
				}
			});

			// uiHelper.CreateSpacer(true, 50);

			toggle = uiHelper.CreateToggle(quickCanvasVAMPresetButtons, false);
			toggle.label = "Quick Canvas Buttons:\nClothing/Appearance Presets";
			toggle.height = 80;

			// Merge presets for VAM default buttons will ALWAYS be enabled now!
			//toggle = uiHelper.CreateToggle(quickCanvasVAMMergePresetsButtons, true);
			//toggle.label = "Quick Canvas Button for VAM:\nUseMergeLoad (M)";
			//toggle.height = 80;

			toggle = uiHelper.CreateToggle(quickCanvasVAMPluginPresetButton, true);
			toggle.label = "Quick Canvas Button:\nMerge Plugin Presets";
			toggle.height = 80;

			// uiHelper.CreateSpacer(true, 80);

			toggle = uiHelper.CreateToggle(quickCanvasCUAMButtons, false);
			toggle.label = "Quick Canvas Buttons for CustomUnityAssetManager\nLoad/Remove Presets";
			toggle.height = 120;

			toggle = uiHelper.CreateToggle(quickCanvasCUAMButtonsCompactLayout, true);
			toggle.label = "Compact Button Layout for CustomUnityAssetManager\nAppend (A) and Remove All (All)";
			toggle.height = 120;

			toggle = uiHelper.CreateToggle(quickCanvasCUAMToggleTriggers, false);
			toggle.label = "Quick Canvas Button for CustomUnityAssetManager\nQuick Toggle Triggers";
			toggle.height = 120;


			uiHelper.CreateSpacer(false, 30);
			uiHelper.CreateSpacer(true, 170);

			btn = uiHelper.CreateButton("Desktop position/layout:", false);
			btn.buttonColor = new Color(1.0f, 0.9f, 0.7f);
			uiHelper.CreateSpacer(true, 45);

			UIDynamicSlider slider = uiHelper.CreateSlider(quickCanvasButtonsXOffset, false);
			slider.labelText.text = "Quick Canvas Buttons\nDesktop Position X";
			slider.valueFormat = "F3";
			slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);

			slider = uiHelper.CreateSlider(quickCanvasButtonsYOffset, false);
			slider.labelText.text = "Quick Canvas Buttons\nDesktop Position Y";
			slider.valueFormat = "F3";
			slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);

			popup = uiHelper.CreateFilterablePopup(quickCanvasButtonsType, true);
			popup.popupPanelHeight = 300f;

			uiHelper.CreateSpacer(false, 30);
			uiHelper.CreateSpacer(true, 170);

			btn = uiHelper.CreateButton("VR position/layout:", false);
			btn.buttonColor = new Color(1.0f, 0.9f, 0.7f);
			uiHelper.CreateSpacer(true, 45);

			slider = uiHelper.CreateSlider(quickCanvasButtonsVRXOffset, false);
			slider.labelText.text = "Quick Canvas Buttons\nVR Position X";
			slider.valueFormat = "F3";
			slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);

			slider = uiHelper.CreateSlider(quickCanvasButtonsVRYOffset, false);
			slider.labelText.text = "Quick Canvas Buttons\nVR Position Y";
			slider.valueFormat = "F3";
			slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);

			slider = uiHelper.CreateSlider(quickCanvasButtonsVRZOffset, false);
			slider.labelText.text = "Quick Canvas Buttons\nVR Position Z";
			slider.valueFormat = "F3";
			slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);

			popup = uiHelper.CreateFilterablePopup(quickCanvasButtonsVRType, true);
			popup.popupPanelHeight = 300f;

			slider = uiHelper.CreateSlider(quickCanvasButtonsVRXRotOffset, true);
			slider.labelText.text = "Quick Canvas Buttons\nVR Rotation X";
			slider.valueFormat = "F1";
			// slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);

			slider = uiHelper.CreateSlider(quickCanvasButtonsVRYRotOffset, true);
			slider.labelText.text = "Quick Canvas Buttons\nVR Rotation Y";
			slider.valueFormat = "F1";
			// slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);


			uiHelper.CreateSpacer(true, 20);
			uiHelper.CreateSpacer(false, 80);

		}

		public void SuggestLoadPluginPreset(string pluginPreset)
        {
			addLoadTriggerCUAMPresetValue.val = "";
			addLoadTriggerPluginPresetValue.val = pluginPreset;
			UpdateAddTriggerPreviewInfoText();
		}

		public void SuggestLoadPreset(string loadPreset)
        {
			addLoadTriggerCUAMPresetValue.val = loadPreset;
			addLoadTriggerPluginPresetValue.val = "";
			
			Helpers.StoreFixedPresetNameSuggestion(loadPreset);
			UpdateAddTriggerPreviewInfoText();
		}

		public void UpdateAddTriggerPreviewInfoText()
        {
			StringBuilder str = new StringBuilder();
			str.AppendLine("<b>Trigger Morph:</b>");
			if (addLoadTriggerMorphSelection.val != "")
            {
				str.AppendLine(addLoadTriggerMorphSelection.val);
            } 
			else
            {
				str.AppendLine("");
            }

			if (addLoadTriggerCUAMPresetValue.val != "")
			{
				str.AppendLine("<b>Load CUAM Preset:</b>");
				str.AppendLine(addLoadTriggerCUAMPresetValue.val);
			} 
			else if (addLoadTriggerPluginPresetValue.val != "")
            {
				str.AppendLine("<b>Load Plugin Preset:</b>");
				str.AppendLine(addLoadTriggerPluginPresetValue.val);
			} 
			else
            {
				str.AppendLine("<b>Load Preset:</b>");
				str.AppendLine("");
			}

			if (addLoadTriggerPersonUidSelection.val != "")
			{
				str.AppendLine("<b>Trigger Person Uid:</b>");
				str.AppendLine(addLoadTriggerPersonUidSelection.val);
			}

			addLoadTriggerPreviewInfo.val = str.ToString();
		}

		public void Reset()
        {
			resetActiveMorphsAfterLoading = true;
		}

		public void Update(Atom personAtom)
        {
			if (enableLoadTriggers.val)
			{
				if (lastRootAtom != personAtom)
                {
					// When root atom changes, delay a few updates before checking auto-loaders.
					lastRootAtom = personAtom;
					autoLoadCheckerCounter = 35;
				}
				if (personAtom != null && !personAtom.destroyed && (autoLoadCheckerCounter++) % 81 == 0) // check each 81 updates, good enough!
				{
					autoLoadCheckerCounter = 1;
					CheckAutoLoaders(personAtom);
				}
			}
			if (optionsDebounceSaveEnabled)
            {
				if (optionsDebounceSaveTime != DateTime.MinValue && DateTime.Now > optionsDebounceSaveTime)
                {
					optionsDebounceSaveTime = DateTime.MinValue;
					optionsDebounceSaveEnabled = false;
					SaveOptions();
				}
            }
		}

		private void ManageAutoLoadTriggersHelp()
        {
			try
			{
				SuperController.singleton.ClearMessages();
				SuperController.LogMessage("");
				SuperController.LogMessage("Manage auto load triggers");
				SuperController.LogMessage("");
				SuperController.LogMessage("An auto load trigger will load a preset based on an active morph on the active selected person atom. The idea is that a unique look will have a unique morph.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* To create an auto load trigger: Select a unique trigger morph for your look and select a preset and press the 'Add/Create Trigger' button. The same morph can have multiple triggers.");
				SuperController.LogMessage("");
				SuperController.LogMessage("Next time you change appearance all presets set up with auto load triggers will be loaded automatically and unloaded when changing again.");
				SuperController.LogMessage("");
				SuperController.LogMessage("If you don't have a unique morph for a look, you can download an empty morph pack and use an empty morph together with your look. Plugins such as MorphMerger can be used to create unique morphs that you can use with your look and auto load triggers.");
				SuperController.LogMessage("Use the *ANY* morph to make the trigger always active on all looks, if you have character details you always want to apply to all your looks, such as an iris reflections or similar.");
				SuperController.LogMessage("");
				SuperController.LogMessage("NOTE: Adding a trigger will automatically load it.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Use the filter/sort options to view all your existing triggers.");
				SuperController.LogMessage("");
				SuperController.singleton.OpenMessageLogPanel();
			}
			catch (Exception e)
			{
				SuperController.LogError(e.ToString());
			}
		}

		private void OptionsHelp()
        {
			try
			{
				SuperController.singleton.ClearMessages();
				SuperController.LogMessage("");
				SuperController.LogMessage("...Scroll down to read everything...");
				SuperController.LogMessage("");
				SuperController.LogMessage("Options");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Auto Select Person From Scene - Keeps the selected person in the scene as the active person in the plugin.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Auto Select Asset Controller - Selects the selected asset controller in the scene automatically so you can grab and move/rotate the controller in the scene without leaving the plugin view.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Delay Physic Properties On Load Preset - Delay setting PhysicLink, PhysicsEnabled and Collision properties on created atoms to prevent crazy physics.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Enable Scaling Modifier - Disable this if you have other plugins that might handle scaling on linked atoms instead.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Update AtomUids In Scripts And Properties - Update any referenced atomUids in scripts and properties to match the created atom's uid in case of duplicate uids.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Enable Prioritized Loading Triggers - If you have two or more triggers using the same trigger morph, personUid will be used to narrow down the hit result to the most prioritized trigger. This allows you to have variations of the same look with different personUid.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Reuse CUA AssetUrl On Load If Same Filename - A feature that reuses existing assetUrls from other loaded CustomUnityAssets when loading a new CustomUnityAssets if the assetUrl uses the same filename for the assetbundle. This is to avoid errors when trying to load the same assetbundle file but from another destination (different .var packages including the same assetbundle for example). This option assumes that same filenames share the same assetbundle file.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Enable Update In Play Mode - Normally you change appearance and clothing for a person while in Edit Mode. This allows the plugin to make a minimal impact while in Play Mode. If you need this plugin to also check for look changes and clothing changes while in Play Mode then you should enable this option.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Ignore Disabled Linked Atoms - When saving presets this will ignore any linked disabled atoms (where On == false).");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Filter: Ignore All Scene Assets - Scene assets will be ignored when 'Loading Presets' and 'Remove All Linked Assets' are used. Click 'Remove All Linked Assets' several times to force removal and ignore this rule.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Filter: Ignore Scene Assets Containing Scripts - Same as Ignore Scene Assets but the asset will be ignored if it contains any script.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Filter: Ignore Scene Assets Marked Hidden - Same as Ignore Scene Assets but the asset will be ignored if it is hidden.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Enable Load Triggers - If load triggers should be enabled. Load triggers automatically load presets based on morph on a character. Here you set up that morph X should load preset Z. it's all managed with the options to the right.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Enable Smart Auto-Load-Trigger Preset Removal - Analyzes scene atoms linked to a person to determine if they belong to an existing preset and should be unloaded when a look is changed. This will allow the plugin to properly remove assets belonging to a look but prevent removing linked assets that belongs to the scene. Requires 'Enable Load Triggers' to be enabled.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Show Deprecated 'Trigger Person Uid' - If you really want to use the deprecated person uid together with triggers.");
				SuperController.LogMessage("");
				SuperController.LogMessage("* Atom Types To Handle - You can add additional atom types here that you want the plugin to handle.");
				SuperController.LogMessage("");
				SuperController.LogMessage("...Scroll to top to read everything...");
				SuperController.singleton.OpenMessageLogPanel();
			}
			catch (Exception e)
			{
				SuperController.LogError(e.ToString());
			}
		}

		private void PluginQuirksHelp()
        {
			try
			{
				SuperController.singleton.ClearMessages();
				SuperController.LogMessage("* QUIRKS:");
				SuperController.LogMessage("");
				SuperController.LogMessage("If you remove a CustomUnityAsset atom before the assetbundle has finish loading, VAM's CustomUnityAssetLoader will lock itself and a hard reset must be performed.");
				SuperController.LogMessage("I couldn't figure out a way to properly see if it's currently loading so this plugin will just prevent you from double-clicking buttons for a short while.");
				SuperController.LogMessage("If you intend to load assetbundles that takes a few seconds to load you must wait for the loading to complete before removing any atoms.");
				SuperController.LogMessage("");
				SuperController.LogMessage("Auto load triggers make a lot of assumtions when clothing item presets should be loaded or unloaded depending on the look/morphs.");
				SuperController.LogMessage("The plugin assumes you're doing manual edits if only one clothing item is enabled/disabled. To avoid false positives from this assumtion when loading clothing presets in certain situations you'll need to double-click when disabling clothing items bound to an auto load trigger.");
				SuperController.LogMessage("");
				SuperController.LogMessage("Plugin triggers will use the pluginLabel to identify when loaded from this plugin and from which source. Therefor, plugins modifying and updating pluginLabels are not supported!");
				SuperController.singleton.OpenMessageLogPanel();
			}
			catch (Exception e)
			{
				SuperController.LogError(e.ToString());
			}
		}

		/// <summary>
		/// Prepare current active morphs when selecting a new person not yet selected. 
		/// This will allow CheckAutoLoaders to properly work depending on how fast the user clicks around in the scene and GUI and load other looks.
		/// </summary>
		/// <param name="atom">Person atom type</param>
		public void PersonSelected(Atom atom)
        {
			if (!atom || atom.destroyed)
				return;

			if (enableLoadTriggers.val && !activeTriggersPerAtom.ContainsKey(atom))
			{
				CheckAutoLoaders(atom);
				autoLoadCheckerCounter = 10;
			}
		}

		/// <summary>
		/// Forces a morph changed check when deselecting a person.
		/// This will allow CheckAutoLoaders to properly work depending on how fast the user clicks around in the scene and GUI and load other looks.
		/// </summary>
		/// <param name="atom"></param>
		public void PersonDeselected(Atom atom)
        {
			if (!atom || atom.destroyed)
				return;

			if (enableLoadTriggers.val && activeTriggersPerAtom.ContainsKey(atom))
			{
				CheckAutoLoaders(atom);
				autoLoadCheckerCounter = 10;
			}
		}

		/// <summary>
        /// Gets the current ActiveTriggerStates for the provided atom. Returns null if no active triggers found.
        /// </summary>
		public AtomTriggerStates GetActiveTriggerStates(Atom atom)
        {
			AtomTriggerStates state = null;
			if (atom != null && activeTriggersPerAtom.TryGetValue(atom, out state))
			{
				return state;
			}
			return null;
		}

		private void CheckAutoLoaders(Atom atom)
		{
			if (atom == null || atom.destroyed)
				return;

			DAZCharacterSelector geometry = atom.GetStorableByID("geometry") as DAZCharacterSelector;
			if (geometry == null || geometry.morphBank1 == null)
			{
				return;
			}

			int loadMode = 1; // 1 ALL , 2 = CLOTHING ONLY
			if (resetActiveMorphsAfterLoading)
			{
				activeTriggersPerAtom.Clear();
				resetActiveMorphsAfterLoading = false;
				// Don't load changes after initial loading
				loadMode = 2;
			}

			AtomTriggerStates state = null;
			if (!activeTriggersPerAtom.TryGetValue(atom, out state))
			{
				state = new AtomTriggerStates();
				activeTriggersPerAtom.Add(atom, state);
				// Don't load changes for a new atom
				loadMode = 2;
			}

			if (autoLoadTriggerPresets.Count == 0 && state.activeTriggers.Count == 0)
			{
				// Nothing to process, nothing to load/unload
				return;
			}

			List<AutoLoadTriggerInfo> triggersToUnload = null;
			List<AutoLoadTriggerInfo> triggersToLoad = null;
			HashSet<string> prioritizedMorphTriggers = new HashSet<string>();

			Dictionary<string, List<DAZMorph>> stateMorphBank = state.GetStateMorphBank(geometry, autoLoadTriggerMorphNames);


			foreach (AutoLoadTriggerInfo loadTrigger in autoLoadTriggerPresets)
			{
				// Find out if this trigger should be currently ACTIVE or NOT.
				bool triggerActive = true;
				// first check personUid.
				triggerActive = (loadTrigger.personUid == "" || loadTrigger.personUid == atom.uid);

				// ANY_MORPH means the trigger is ALWAYS active no matter the person's morphs and can't be overriden by prioritized morphs.

				// check if any prioritized trigger has been found with the same morphName and a personUid already, any triggers after with no personUid but same morph should be ignored.
				if (loadTrigger.personUid == "" && loadTrigger.morphName != ANY_MORPH && enablePrioritizedLoadingTriggers.val && prioritizedMorphTriggers.Contains(loadTrigger.morphName))
				{
					triggerActive = false;
				}

				// continue to check morphs
				if (triggerActive && loadTrigger.morphName != "")
                {
					if (loadTrigger.morphName != ANY_MORPH)
					{
						// morph.active || morph.startValue != morph.appliedValue || morph.morphValue != morph.jsonFloat.defaultVal

						// triggerActive = autoLoadTriggerMorphs.Any(morph => { return (morph.displayName == loadTrigger.morphName && (morph.active || morph.morphValue != morph.jsonFloat.defaultVal)); });
						List<DAZMorph> morphs = null;
						triggerActive = stateMorphBank.TryGetValue(loadTrigger.morphName, out morphs) && morphs.Any((morph) => { return (morph.active || morph.morphValue != morph.jsonFloat.defaultVal); });
					}
				}

				if (triggerActive && loadTrigger.personUid != "" && loadTrigger.morphName != "")
                {
					prioritizedMorphTriggers.Add(loadTrigger.morphName);
				}

				// Track active triggers for person, any that needs to be loaded/unloaded!?
				if (!triggerActive && state.activeTriggers.Contains(loadTrigger))
				{
					state.activeTriggers.Remove(loadTrigger);
					// SuperController.LogMessage("Removing " + loadTrigger.morphName + " => " + loadTrigger.presetName);

					if (smartRemovalOfPresets.val)
					{
						if (triggersToUnload == null)
							triggersToUnload = new List<AutoLoadTriggerInfo>();
						triggersToUnload.Add(loadTrigger);
					}
				} 
				
				if (triggerActive && !state.activeTriggers.Contains(loadTrigger))
				{
					state.activeTriggers.Add(loadTrigger);
					// SuperController.LogMessage("Adding " + loadTrigger.morphName + " => " + loadTrigger.presetName);

					if (triggersToLoad == null)
						triggersToLoad = new List<AutoLoadTriggerInfo>();
					triggersToLoad.Add(loadTrigger);
				}
			}

			// Activate triggers
			if (triggersToLoad != null || triggersToUnload != null)
			{
				if (CustomUnityAssetManager.DEBUG)
				{
					SuperController.LogMessage("PersonUid/Morph change detected with load mode " + loadMode.ToString());
					string debug = "";
					if (triggersToLoad != null)
					{
						debug += (" 'Presets to LOAD': " + string.Join(", ", triggersToLoad.Select((x) => { return x.presetName; }).ToArray()));
					}
					if (triggersToUnload != null)
					{
						debug += (" 'Presets to UNLOAD': " + string.Join(", ", triggersToUnload.Select((x) => { return x.presetName; }).ToArray()));
					}
					if (debug != "")
					{
						SuperController.LogMessage(debug);
					}
				}

				core.AutoLoaderTriggered(atom, loadMode,
										 triggersToLoad?.Select((x) => { return x; }).ToList(), 
										 triggersToUnload?.Select((x) => { return x; }).ToList());

				UpdateAutoLoaderPresetsListForActiveTriggers();
			}
		}

		protected void SelectedOptionsToggleChanged(bool val)
		{
			SaveOptions();
		}

		public void SaveOptions(float debounceSeconds = 0f)
		{
			if (debounceSeconds > 0f)
            {
				optionsDebounceSaveEnabled = true;
				optionsDebounceSaveTime = DateTime.Now.AddSeconds(debounceSeconds);
				return;
            }
			optionsDebounceSaveEnabled = false;
			StringBuilder sb = new StringBuilder();
			sb.AppendLine("{");
			sb.AppendLine("  \"id\" : \"Options\",");
			sb.AppendLine("  \"autoSelectPerson\" : \"" + (autoSelectPersonFromScene.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"autoSelectAssetController\" : \"" + (autoSelectAssetController.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"delayPhysicsProperties\" : \"" + (delayPhysicsProperties.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"enableScalingModifier\" : \"" + (enableScalingModifier.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"preservePersonScaleOnChange\" : \"" + (preservePersonScaleOnChange.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"enableLoadTriggers\" : \"" + (enableLoadTriggers.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"ignoreAllSceneAssets\" : \"" + (ignoreAllSceneAssets.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"ignoreSceneAsssetsContainingScripts\" : \"" + (ignoreSceneAssetsContainingScripts.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"ignoreSceneAssetsMarkedHidden\" : \"" + (ignoreSceneAssetsMarkedHidden.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"smartRemovalOfPresets\" : \"" + (smartRemovalOfPresets.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"cuabVerbosePluginLogs\" : \"" + (verbosePluginLogs.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"showLegacyPersonUidTriggerSelector\" : \"" + (showLegacyPersonUidTriggerSelector.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"updateScriptAtomUids\" : \"" + (updateScriptAtomUids.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"enablePrioritizedLoadingTriggers\" : \"" + (enablePrioritizedLoadingTriggers.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"enableCUAReuseOnLoad\" : \"" + (enableCUAReuseOnLoad.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"enablePlayModeUpdates\" : \"" + (enablePlayModeUpdates.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"ignoreDisabledLinkedAtoms\" : \"" + (ignoreDisabledLinkedAtoms.val ? "true" : "false") + "\",");

			sb.AppendLine("  \"quickCanvasCUAMButtons\" : \"" + (quickCanvasCUAMButtons.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"quickCanvasVAMPresetButtons\" : \"" + (quickCanvasVAMPresetButtons.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"quickCanvasVAMPluginPresetButton\" : \"" + (quickCanvasVAMPluginPresetButton.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"quickCanvasVAMClothingMergePresetsButton\" : \"" + (quickCanvasVAMMergePresetsButtons.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"quickCanvasCUAMButtonsCompactLayout\" : \"" + (quickCanvasCUAMButtonsCompactLayout.val ? "true" : "false") + "\",");
			sb.AppendLine("  \"quickCanvasCUAMToggleTriggers\" : \"" + (quickCanvasCUAMToggleTriggers.val ? "true" : "false") + "\",");

			sb.AppendLine("  \"quickCanvasButtonsType\" : \"" + quickCanvasButtonsType.val + "\",");
			sb.AppendLine("  \"quickCanvasButtonsXOffset\" : \"" + Math.Round(quickCanvasButtonsXOffset.val, 4).ToString() + "\",");
			sb.AppendLine("  \"quickCanvasButtonsYOffset\" : \"" + Math.Round(quickCanvasButtonsYOffset.val, 4).ToString() + "\",");

			sb.AppendLine("  \"quickCanvasButtonsVRType\" : \"" + quickCanvasButtonsVRType.val + "\",");
			sb.AppendLine("  \"quickCanvasButtonsVRXOffset\" : \"" + Math.Round(quickCanvasButtonsVRXOffset.val, 4).ToString() + "\",");
			sb.AppendLine("  \"quickCanvasButtonsVRYOffset\" : \"" + Math.Round(quickCanvasButtonsVRYOffset.val, 4).ToString() + "\",");
			sb.AppendLine("  \"quickCanvasButtonsVRZOffset\" : \"" + Math.Round(quickCanvasButtonsVRZOffset.val, 4).ToString() + "\",");
			sb.AppendLine("  \"quickCanvasButtonsVRXRotOffset\" : \"" + Math.Round(quickCanvasButtonsVRXRotOffset.val, 4).ToString() + "\",");
			sb.AppendLine("  \"quickCanvasButtonsVRYRotOffset\" : \"" + Math.Round(quickCanvasButtonsVRYRotOffset.val, 4).ToString() + "\",");

			sb.AppendLine("  \"includedAtomTypes\" : \"" + (string.Join(",", includedAtomTypes.ToArray())) + "\",");
			sb.AppendLine("  \"autoMorphLoaders\" : [");
			AutoLoadTriggerInfo[] loadTriggers = autoLoadTriggerPresets.Where((x) => { return !x.external; }).ToArray();
			for (int i = 0; i < loadTriggers.Length; i++)
			{
				sb.AppendLine("  {");
				sb.AppendLine("  \"id\" : \"" + loadTriggers[i].morphName + "\",");
				sb.AppendLine("  \"personUid\" : \"" + loadTriggers[i].personUid + "\",");
				sb.AppendLine("  \"preset\" : \"" + loadTriggers[i].presetName + "\",");
				sb.AppendLine("  \"type\" : \"" + loadTriggers[i].presetType + "\",");
				sb.AppendLine("  }" + (i < loadTriggers.Length - 1 ? "," : ""));
			}
			sb.AppendLine("  ]");
			sb.AppendLine("}");
			JSONClass mainjc = JSONNode.Parse(sb.ToString()).AsObject;
			try
			{
				// Save to new plugin folder Saves\\PluginData\\Blazedust\\CUAManager\\options.json
				// core.pluginPath.PluginPresetPath <-- LEGACY path
				// But now always store to the new path (and hopefully avoid confirm overwrite dialog)
				SuperController.singleton.SaveJSON(mainjc, core.pluginPath.PluginDataFolder + "/options.json");
			}
			catch (Exception e)
			{
				SuperController.LogError(e.ToString());
			}
		}

		/// <summary>
		/// Reads options.json from the PluginDataFolder (v.1.20.77 in VAM) first and secondly from the legacy folder.
		/// </summary>
		public JSONNode GetOptionsJSONFromFile()
		{
			JSONNode jsoptions = null;
			string[] files = SuperController.singleton.GetFilesAtPath(core.pluginPath.PluginDataFolder, "options.json");
			if (files.Length > 0)
            {
				jsoptions = SuperController.singleton.LoadJSON(core.pluginPath.PluginDataFolder + "/options.json");
				if (jsoptions != null)
					return jsoptions;
			}
			files = SuperController.singleton.GetFilesAtPath(core.pluginPath.PluginPresetPath, "options.json");
			if (files.Length > 0)
			{
				jsoptions = SuperController.singleton.LoadJSON(core.pluginPath.PluginPresetPath + "/options.json");
			}
			return jsoptions;
		}

		public void LoadOptions()
        {
			try
			{
				includedAtomTypes.Clear();
				autoLoadTriggerPresets.Clear();
				autoLoadTriggerMorphNames.Clear();

				JSONNode jsoptions = GetOptionsJSONFromFile();
				if (jsoptions != null)
				{
					autoSelectPersonFromScene.valNoCallback = (jsoptions["autoSelectPerson"] == null || jsoptions["autoSelectPerson"].Value == "true");
					autoSelectAssetController.valNoCallback = (jsoptions["autoSelectAssetController"] == null || jsoptions["autoSelectAssetController"].Value == "true");
					delayPhysicsProperties.valNoCallback = (jsoptions["delayPhysicsProperties"] == null || jsoptions["delayPhysicsProperties"].Value == "true");
					enableScalingModifier.valNoCallback = (jsoptions["enableScalingModifier"] == null || jsoptions["enableScalingModifier"].Value == "true");
					preservePersonScaleOnChange.valNoCallback = (jsoptions["preservePersonScaleOnChange"] == null || jsoptions["preservePersonScaleOnChange"].Value == "true");
					enableLoadTriggers.valNoCallback = (jsoptions["enableLoadTriggers"] == null || jsoptions["enableLoadTriggers"].Value == "true");
					ignoreAllSceneAssets.valNoCallback = (jsoptions["ignoreAllSceneAssets"] == null || jsoptions["ignoreAllSceneAssets"].Value == "true");
					ignoreSceneAssetsContainingScripts.valNoCallback = (jsoptions["ignoreSceneAsssetsContainingScripts"] == null || jsoptions["ignoreSceneAsssetsContainingScripts"].Value == "true");
					ignoreSceneAssetsMarkedHidden.valNoCallback = (jsoptions["ignoreSceneAssetsMarkedHidden"] == null || jsoptions["ignoreSceneAssetsMarkedHidden"].Value == "true");
					smartRemovalOfPresets.valNoCallback = (jsoptions["smartRemovalOfPresets"] == null || jsoptions["smartRemovalOfPresets"].Value == "true");
					showLegacyPersonUidTriggerSelector.valNoCallback = (jsoptions["showLegacyPersonUidTriggerSelector"] != null && jsoptions["showLegacyPersonUidTriggerSelector"].Value == "true");
					verbosePluginLogs.valNoCallback = (jsoptions["cuabVerbosePluginLogs"] == null || jsoptions["cuabVerbosePluginLogs"].Value == "true");
					updateScriptAtomUids.valNoCallback = (jsoptions["updateScriptAtomUids"] == null || jsoptions["updateScriptAtomUids"].Value == "true");
					enablePrioritizedLoadingTriggers.valNoCallback = (jsoptions["enablePrioritizedLoadingTriggers"] != null && jsoptions["enablePrioritizedLoadingTriggers"].Value == "true");
					enableCUAReuseOnLoad.valNoCallback = (jsoptions["enableCUAReuseOnLoad"] == null || jsoptions["enableCUAReuseOnLoad"].Value == "true");
					enablePlayModeUpdates.valNoCallback = (jsoptions["enablePlayModeUpdates"] != null && jsoptions["enablePlayModeUpdates"].Value == "true");
					ignoreDisabledLinkedAtoms.valNoCallback = (jsoptions["ignoreDisabledLinkedAtoms"] == null || jsoptions["ignoreDisabledLinkedAtoms"].Value == "true");

					quickCanvasCUAMButtons.valNoCallback = (jsoptions["quickCanvasCUAMButtons"] == null || jsoptions["quickCanvasCUAMButtons"].Value == "true");
					quickCanvasVAMPresetButtons.valNoCallback = (jsoptions["quickCanvasVAMPresetButtons"] == null || jsoptions["quickCanvasVAMPresetButtons"].Value == "true");
					quickCanvasVAMPluginPresetButton.valNoCallback = (jsoptions["quickCanvasVAMPluginPresetButton"] == null || jsoptions["quickCanvasVAMPluginPresetButton"].Value == "true");
					quickCanvasVAMMergePresetsButtons.valNoCallback = (jsoptions["quickCanvasVAMClothingMergePresetsButton"] != null && jsoptions["quickCanvasVAMClothingMergePresetsButton"].Value == "true");
					quickCanvasCUAMButtonsCompactLayout.valNoCallback = (jsoptions["quickCanvasCUAMButtonsCompactLayout"] != null && jsoptions["quickCanvasCUAMButtonsCompactLayout"].Value == "true");
					quickCanvasCUAMToggleTriggers.valNoCallback = (jsoptions["quickCanvasCUAMToggleTriggers"] != null && jsoptions["quickCanvasCUAMToggleTriggers"].Value == "true");

					quickCanvasButtonsType.valNoCallback = (jsoptions["quickCanvasButtonsType"] != null ? jsoptions["quickCanvasButtonsType"].Value : QUICK_CANVAS_BUTTON_LAYOUT_STACKED);
					float floatValue = 0f;
					string stringValue = (jsoptions["quickCanvasButtonsXOffset"] != null ? jsoptions["quickCanvasButtonsXOffset"].Value : "");
					if (float.TryParse(stringValue, out floatValue)) { quickCanvasButtonsXOffset.valNoCallback = floatValue; }
					stringValue = (jsoptions["quickCanvasButtonsYOffset"] != null ? jsoptions["quickCanvasButtonsYOffset"].Value : "");
					if (float.TryParse(stringValue, out floatValue)) { quickCanvasButtonsYOffset.valNoCallback = floatValue; }

					quickCanvasButtonsVRType.valNoCallback = (jsoptions["quickCanvasButtonsVRType"] != null ? jsoptions["quickCanvasButtonsVRType"].Value : QUICK_CANVAS_BUTTON_LAYOUT_HORIZONTAL);
					stringValue = (jsoptions["quickCanvasButtonsVRXOffset"] != null ? jsoptions["quickCanvasButtonsVRXOffset"].Value : "");
					if (float.TryParse(stringValue, out floatValue)) { quickCanvasButtonsVRXOffset.valNoCallback = floatValue; }
					stringValue = (jsoptions["quickCanvasButtonsVRYOffset"] != null ? jsoptions["quickCanvasButtonsVRYOffset"].Value : "");
					if (float.TryParse(stringValue, out floatValue)) { quickCanvasButtonsVRYOffset.valNoCallback = floatValue; }
					stringValue = (jsoptions["quickCanvasButtonsVRZOffset"] != null ? jsoptions["quickCanvasButtonsVRZOffset"].Value : "");
					if (float.TryParse(stringValue, out floatValue)) { quickCanvasButtonsVRZOffset.valNoCallback = floatValue; }

					stringValue = (jsoptions["quickCanvasButtonsVRXRotOffset"] != null ? jsoptions["quickCanvasButtonsVRXRotOffset"].Value : "");
					if (float.TryParse(stringValue, out floatValue)) { quickCanvasButtonsVRXRotOffset.valNoCallback = floatValue; }
					stringValue = (jsoptions["quickCanvasButtonsVRYRotOffset"] != null ? jsoptions["quickCanvasButtonsVRYRotOffset"].Value : "");
					if (float.TryParse(stringValue, out floatValue)) { quickCanvasButtonsVRYRotOffset.valNoCallback = floatValue; }

					if (jsoptions["includedAtomTypes"] != null)
					{
						string values = jsoptions["includedAtomTypes"].Value;
						string[] atomTypes = values.Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries);
						includedAtomTypes.AddRange(atomTypes);
					}

					autoLoadTriggerPresets.Clear();
					autoLoadTriggerMorphNames.Clear();
					if (jsoptions["autoMorphLoaders"] != null)
					{
						foreach (JSONNode n in jsoptions["autoMorphLoaders"].AsArray.Childs)
						{
							AutoLoadTriggerInfo loadTrigger = new AutoLoadTriggerInfo(n["personUid"], n["id"], n["preset"], "", n["type"]);
							autoLoadTriggerPresets.Add(loadTrigger);
							autoLoadTriggerMorphNames.Add(loadTrigger.morphName);
							// SuperController.LogMessage(loadTrigger.ToString());
						}
					}
				}

				// load external autoMorphLoaders option files
				jsoptions = null;
				string[] files = SuperController.singleton.GetFilesAtPath(core.pluginPath.PluginPresetPath, "*.json");
				files = files.Union(SuperController.singleton.GetFilesAtPath(core.pluginPath.PluginDataFolder, "*.json")).ToArray();
				if (files.Length > 0)
                {
					foreach(string file in files)
                    {
						string fileName = Path.GetFileNameAndExt(file);
						if (!fileName.StartsWith("options_", StringComparison.InvariantCultureIgnoreCase))
                        {
							continue;
                        }

						try
                        {
							jsoptions = SuperController.singleton.LoadJSON(core.pluginPath.PluginPresetPath + "/" + fileName);
							if (jsoptions["includedAtomTypes"] != null)
							{
								string values = jsoptions["includedAtomTypes"].Value;
								string[] atomTypes = values.Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries);
								includedAtomTypes.AddRange(atomTypes);
							}

							if (jsoptions["autoMorphLoaders"] != null)
							{
								foreach (JSONNode n in jsoptions["autoMorphLoaders"].AsArray.Childs)
								{
									AutoLoadTriggerInfo loadTrigger = new AutoLoadTriggerInfo(n["personUid"], n["id"], n["preset"], fileName, n["type"]);
									autoLoadTriggerPresets.Add(loadTrigger);
									autoLoadTriggerMorphNames.Add(loadTrigger.morphName);
									// SuperController.LogMessage(loadTrigger.ToString());
								}
							}
						}
						catch (Exception e)
						{
							// Ignore, file might not exist or loading values failed...
							SuperController.LogError("");
							SuperController.LogError("Failed to load option file " + file);
							SuperController.LogError(e.ToString());
						}
					}
				}
			}
			catch (Exception e)
			{
				// Ignore, file might not exist or loading values failed...
				SuperController.LogError(e.ToString());
			}

			// Remove duplicates from includedAtomTypes
			includedAtomTypes = includedAtomTypes.Distinct().ToList(); // easy way of doing it, will work!
			// Force "CustomUnityAsset" to be included, plugin designed for that
			if (!includedAtomTypes.Contains("CustomUnityAsset"))
            {
				includedAtomTypes.Insert(0, "CustomUnityAsset");
			}

			// Remove duplicates from autoLoaderMorphPresets
			HashSet<string> keys = new HashSet<string>();
			for (int i = 0; i < autoLoadTriggerPresets.Count; i++)
			{
				if (!keys.Add(autoLoadTriggerPresets[i].presetName + ":" + autoLoadTriggerPresets[i].morphName + ":" + autoLoadTriggerPresets[i].personUid + ":" + autoLoadTriggerPresets[i].presetType))
                {
					// duplicate
					// SuperController.LogMessage("Removing duplicate: " + autoLoaderMorphPresets[i].ToString());
					autoLoadTriggerPresets.RemoveAt(i);
					i--;
				}
			}
			SortAutoLoadTriggerPresets();
			
			UpdateAutoLoaderPresetsList();
			UpdateIncludedAtomTypesList();
		}

		private void SortAutoLoadTriggerPresets()
        {
			// let personUid triggers come before others, this will be important to make certain assumtions in the auto-loader routine check optimized
			autoLoadTriggerPresets.Sort((a, b) => {
				int c = string.Compare(b.personUid, a.personUid); // key 1
				if (c == 0)
                {
					int c2 = string.Compare(a.morphName, b.morphName); // key 2
					if (c2 == 0)
                    {
						int c3 = string.Compare(a.presetName, b.presetName); // key 3
						if (c3 == 0)
                        {
							int c4 = string.Compare(a.externalFile, b.externalFile); // key 4
							if (c4 == 0)
                            {
								// key 5: HashKey
								// This SHOULD ensure that the sorting remains THE SAME even if same morph is used over multiple presets
								if (a.GetHashCode() > b.GetHashCode())
                                {
									return -1;
                                } 
								else
                                {
									return 1;
                                }
                            }
							return c4;
						}
						return c3;
					}
					return c2;
                }
				return c;
			});
			// SuperController.LogMessage("autoLoadTriggerPresets: " + string.Join(", ", autoLoadTriggerPresets.Select((x) => { return x.personUid + ":" + x.morphName; }).ToArray()));
		}

		public void UpdateAutoLoaderPresetsListForActiveTriggers()
        {
			if (addLoadTriggerSorting != null && addLoadTriggerSorting.val == TRIGGER_SORT_ACTIVE_TRIGGERS)
            {
				UpdateAutoLoaderPresetsList();
			}
        }
		public void UpdateAutoLoaderPresetsList()
		{
			UpdateAutoLoaderPresetsListText();

			AutoLoadTriggerInfo selectedTriggerInfo = GetSelectedAutoLoadTriggerInfo();

			RefreshRemoveLoadSelectedTriggersList();

			core.DoNextFrame(() => {
				if (!SetSelectedAutoLoadTriggerInfo(selectedTriggerInfo))
                {
					removeLoadTriggersSelection.valNoCallback = "";
				}
			});
		}

		public void UpdateAutoLoaderPresetsListText()
        {
			StringBuilder text = new StringBuilder();
			text.Append("<b>Auto load preset triggers:</b>\n\n");
			if (!enableLoadTriggers.val)
			{
				text.Append("Load Triggers DISABLED from options.\n\n");
			}

			IEnumerable<AutoLoadTriggerListInfo> triggers = autoLoadTriggerPresets.Select((x, i) => { return new AutoLoadTriggerListInfo(x, i); });
			if (addLoadTriggerSorting != null)
			{
				switch (addLoadTriggerSorting.val)
				{
					case TRIGGER_SORT_TRIGGER_DEFAULT:
						/* default sorting as is */
						break;
					case TRIGGER_SORT_MORPH_NAME:
						triggers = triggers.OrderBy(t => t.Info.morphName);
						break;
					case TRIGGER_SORT_PRESET_NAME:
						triggers = triggers.OrderBy(t => t.Info.presetName);
						break;
					case TRIGGER_SORT_PERSON_UID:
						triggers = triggers.Where(t => t.Info.personUid != null && t.Info.personUid != "").OrderBy(t => t.Info.personUid);
						if (text != null)
						{
							text.Append(string.Format("[Filter] Showing {0}/{1}\n", triggers.Count(), autoLoadTriggerPresets.Count));
							text.Append("Triggers with PersonUid set:\n\n");
						}
						break;
					case TRIGGER_SORT_ACTIVE_TRIGGERS:

						AtomTriggerStates activeStates = GetActiveTriggerStates(core.rootAtom);
						if (activeStates != null && activeStates.activeTriggers != null && activeStates.activeTriggers.Count > 0)
						{
							triggers = triggers.Where(x => activeStates.activeTriggers.Any(t => t.Equals(x.Info)));
							if (text != null)
							{
								text.Append(string.Format("[Filter] Showing {0}/{1}\n", triggers.Count(), autoLoadTriggerPresets.Count));
								text.Append("Active triggers for " + core.rootAtom.uid + ":\n\n");
							}
						}
						else
						{
							triggers = null;
							if (text != null)
							{
								text.Append(string.Format("[Filter] Showing {0}/{1}\n", 0, autoLoadTriggerPresets.Count));
								if (core.rootAtom == null)
								{
									text.Append("No active person selected. No active triggers.\n\n");
								}
								else
								{
									text.Append("No current active triggers for " + core.rootAtom.uid + ".\n\n");
								}
							}
						}

						break;
				}
			}

			autoLoadTriggerPresetsFiltered.Clear();
			if (triggers != null)
			{
				autoLoadTriggerPresetsFiltered = triggers.ToList();
				int indexRange = (GetTriggerPageSelectedZeroIndexed() * 100);
				int index = 0;
				foreach (AutoLoadTriggerListInfo x in triggers)
				{
					if (index >= indexRange && index < indexRange + 100)
					{
						text.Append("<b>Trigger " + x.Index.ToString() + ":</b>\n" + (x.Info.personUid != "" ? "personUid: " + x.Info.personUid + "\n" : "") + (x.Info.morphName != "" ? "morph: " + x.Info.morphName + "\n" : "") + "preset: " + x.Info.presetName + "\ntype: " + x.Info.presetType + (!x.Info.external ? "" : "\nfile: " + x.Info.externalFile));
						text.Append("\n\n");
					}
					index += 1;
				}
			}

			if (triggers != null && triggers.Count() > 100)
            {
				text.Append("Showing page " + autoLoaderPresetsListPageSelector.val);
            }

			text.Append("\n\n");

			autoLoaderPresetsList.val = text.ToString();
		}

		public List<string> GetTriggerPageOptions()
		{
			List<string> lst = new List<string>();
			if (autoLoadTriggerPresets == null)
			{
				lst.Add("1/1");
			} 
			else
            {
				int totPages = (autoLoadTriggerPresets.Count / 100) + 1;
				for (int i = 0; i < (int)(autoLoadTriggerPresets.Count / 100) + 1; i++)
                {
					int h = i * 100;
					int from = h;
					int to = h + 99;
					//lst.Add(string.Format("{0}-{1}", from, to));
					lst.Add((i + 1).ToString() + "/" + totPages.ToString()); // + " (" + from.ToString() + "-" + to.ToString() + ")");
				}
			}
			if (lst.Count == 0)
            {
				lst.Add("1/1");
			}
			return lst;
		}

		public int GetTriggerPageSelectedZeroIndexed()
        {
			if (autoLoaderPresetsListPageSelector == null)
				return 0;

			string s = autoLoaderPresetsListPageSelector.val;
			if (s == null || s == "")
				return 0;

			if (s.IndexOf("/") > 0)
			{
				s = s.Substring(0, s.IndexOf("/"));
				if (s == "")
					return 0;
			}
			//if (s.IndexOf("-") > 0)
			//{
			//	s = s.Substring(0, s.IndexOf("-"));
			//	if (s == "")
			//		return 0;
			//}

			int r = 0;
			// return page as zero-indexed
			if (int.TryParse(s, out r))
				return r - 1;

			return 0;
		}

		public void AutoLoaderPresetsListPageSelectorUpdated()
        {
			UpdateAutoLoaderPresetsListText();
		}

		public void RefreshRemoveLoadSelectedTriggersList()
		{
			// Refresh trigger selection
			List<string> triggerOptions = GetLoadTriggerNames(true);
			
			// Hack to make the Selected Trigger list only shown triggers based on filter
			List<int> showIndexes = null;
			if (addLoadTriggerSorting != null)
			{
				switch (addLoadTriggerSorting.val)
				{
					case TRIGGER_SORT_TRIGGER_DEFAULT:
						break;
					case TRIGGER_SORT_MORPH_NAME:
						break;
					case TRIGGER_SORT_PRESET_NAME:
						break;
					case TRIGGER_SORT_PERSON_UID:
						{
							IEnumerable<AutoLoadTriggerListInfo> triggers = autoLoadTriggerPresets.Select((x, i) => { return new AutoLoadTriggerListInfo(x, i); });
							showIndexes = new List<int>(triggers.Where(t => t.Info.personUid != null && t.Info.personUid != "").Select(x => x.Index));
						}
						break;
					case TRIGGER_SORT_ACTIVE_TRIGGERS:
						{
							AtomTriggerStates activeStates = GetActiveTriggerStates(core.rootAtom);
							if (activeStates != null && activeStates.activeTriggers != null && activeStates.activeTriggers.Count > 0)
							{
								IEnumerable<AutoLoadTriggerListInfo> triggers = autoLoadTriggerPresets.Select((x, i) => { return new AutoLoadTriggerListInfo(x, i); });
								showIndexes = new List<int>(triggers.Where(x => activeStates.activeTriggers.Any(t => t.Equals(x.Info))).Select(x => x.Index));
							}
						}
						break;
				}
			}

			if (showIndexes != null)
            {
				triggerOptions = new List<string>(showIndexes.Select(x => "Trigger " + x.ToString()));
				triggerOptions.Insert(0, "");
			}
			
			removeLoadTriggersSelection.choices = triggerOptions;
			//	removeLoadTriggersSelection.val = "";
			
		}

		public void SelectTriggerLoadPluginPresetUsingFileBrowser()
        {
			string storeDirectory = "Custom\\Atom\\Person\\Plugins";
			List<MVR.FileManagementSecure.ShortCut> shortcuts = MVR.FileManagementSecure.FileManagerSecure.GetShortCutsForDirectory(storeDirectory, true, true, true, true);
			SuperController.singleton.GetMediaPathDialog((string presetFileName) =>
			{
				// selected something?
				if (presetFileName != null && presetFileName != "")
				{
					SuggestLoadPluginPreset(presetFileName);
				}
			}, "vap", storeDirectory, false, true, false, "Preset_", false, shortcuts);
			
		}

		public void SelectTriggerLoadCUAMPresetUsingFileBrowser()
        {
			string defaultPath = SuperController.singleton.fileBrowserUI.defaultPath;
			SuperController.singleton.fileBrowserUI.defaultPath = core.pluginPath.PluginPresetPath;
			SuperController.singleton.fileBrowserUI.ClearCurrentPath();
			SuperController.singleton.fileBrowserUI.GotoDirectory(core.pluginPath.PluginPresetPath);
			SuperController.singleton.fileBrowserUI.SetTitle("Select Load Preset");
			SuperController.singleton.fileBrowserUI.SetTextEntry(false);

			SuperController.singleton.fileBrowserUI.Show((path) =>
			{
				SuperController.singleton.fileBrowserUI.defaultPath = defaultPath;

				if (path != "")
				{
					// map out the preset name from the selected file and suggest it
					string presetName = core.pluginPath.GetPresetNameFromFileSelection(path);
					SuggestLoadPreset(presetName);
				}
			});
		}

		public void UpdateIncludedAtomTypesList()
        {
			includedAtomTypesList.val = "Atom types to handle (assets):\n\n" + string.Join("\n", includedAtomTypes.ToArray());
        }

		public const string PRESET_CUAM = "CUAMPreset";
		public const string PRESET_PLUGIN = "PluginPreset";

		public AutoLoadTriggerInfo AddLoadTrigger(string personUid, string morphName, string presetName, string presetType)
		{
			if (personUid == null)
				personUid = "";
			if (morphName == null)
				morphName = "";
			if (presetName == null)
				presetName = "";
			if (presetType == null)
				presetType = "";

			if (presetName == null || presetType == null || personUid == "" && morphName == "")
            {
				// nothing to add
				return null;
            }

			// Don't add duplicates.
			AutoLoadTriggerInfo newLoadTrigger = null;
			if (autoLoadTriggerPresets.FirstOrDefault((x) => { return x.personUid == personUid && x.morphName == morphName && x.presetName == presetName && x.presetType == presetType; }) != null)
			{
				SuperController.LogMessage("Trigger already added. Skipping duplicate.");
			}
			else
			{
				ScanPersonsInScene();

				// Log("Trigger added.");
				newLoadTrigger = new AutoLoadTriggerInfo(personUid, morphName, presetName, "", presetType);
				autoLoadTriggerPresets.Add(newLoadTrigger);
				autoLoadTriggerMorphNames.Add(morphName);
				SortAutoLoadTriggerPresets();
				SelectedOptionsToggleChanged(true); // trigger save options

				bool coreRootAtomHandled = false;

				// Update any active triggers
				foreach (KeyValuePair<Atom, AtomTriggerStates> pair in activeTriggersPerAtom)
				{
					pair.Value.AutoLoadTriggerAdded(newLoadTrigger);

					// Make sure to unload any matching CustomUnityAsset presets when creating a trigger, as the newly added trigger will soon load in CustomUnityAssets again.
					if (enableLoadTriggers.val && pair.Key != null && !pair.Key.destroyed)
					{
						coreRootAtomHandled |= (pair.Key == core.rootAtom);
						core.UnloadPreset(pair.Key, presetName, true, false);
					}
				}
				// UpdateAutoLoaderPresetsListForActiveTriggers(); // UpdateAutoLoaderPresetsList fixes this below

				if (!coreRootAtomHandled)
				{
					// Make sure to unload any matching CustomUnityAsset presets when creating a trigger, as the newly added trigger will soon load in CustomUnityAssets again.
					if (enableLoadTriggers.val && core.rootAtom != null && !core.rootAtom.destroyed)
					{
						core.UnloadPreset(core.rootAtom, presetName, true, false);
					}
				}

				//// Force trigger load on the newly saved trigger - avoid this as the auto-load-trigger routine will active the trigger shortly
				//if (enableLoadTriggers.val && core.rootAtom != null && !core.rootAtom.destroyed)
				//{
				//	core.AutoLoaderTriggered(core.rootAtom, 1, new List<AutoLoadTriggerInfo>(new AutoLoadTriggerInfo[] { newLoadTrigger }), null);
				//}

				ScanPersonsInScene();

			}
			UpdateAutoLoaderPresetsList();
			return newLoadTrigger;
		}

		public AutoLoadTriggerInfo GetSelectedAutoLoadTriggerInfo()
		{
			if (removeLoadTriggersSelection != null && removeLoadTriggersSelection.val != "")
			{
				int index = GetLoadTriggerNames(false).IndexOf(removeLoadTriggersSelection.val);
				if (index >= 0)
				{
					return autoLoadTriggerPresets[index];
				}
			}
			return null;
		}

		public bool SetSelectedAutoLoadTriggerInfo(AutoLoadTriggerInfo value)
		{
			if (removeLoadTriggersSelection != null && value != null)
			{
				AutoLoadTriggerListInfo listItem = autoLoadTriggerPresetsFiltered.FirstOrDefault(x => x.Info.Equals(value));
				if (listItem != null)
				{
					removeLoadTriggersSelection.valNoCallback = "Trigger " + listItem.Index.ToString();
					return true;
				}
			}
			return false;
		}

		// "Scan" persons in the scene for active triggers. Intended to be run before/after adding/removing triggers.
		public void ScanPersonsInScene()
        {
			foreach (Atom personAtom in SuperController.singleton.GetAtoms()
				.Where((atom) =>
				{
					return (atom.type == "Person");
				})) 
			{
				bool prevValue = resetActiveMorphsAfterLoading; // keep this intact ... meeh!
				CheckAutoLoaders(personAtom);
				resetActiveMorphsAfterLoading = prevValue;
			}
		}

		public void RemoveLoadTriggerName(string name)
		{
			int index = GetLoadTriggerNames(false).IndexOf(name);
			if (index >= 0)
			{
				AutoLoadTriggerInfo loadTrigger = autoLoadTriggerPresets[index];
				if (loadTrigger.external)
				{
					SuperController.LogError("This trigger is loaded from an external options file (" + loadTrigger.externalFile + "). Triggers loaded from external option files must be removed manually from the plugin preset folder and plugin reloaded.");
				}
				else
				{
					ScanPersonsInScene();

					autoLoadTriggerPresets.RemoveAt(index);

					// Only remove morphNames from trigger if there are no other loadTriggers using this morphName.
					if (!autoLoadTriggerPresets.Any(x => x.morphName == loadTrigger.morphName))
					{
						autoLoadTriggerMorphNames.Remove(loadTrigger.morphName);
					}

					SortAutoLoadTriggerPresets();
					SelectedOptionsToggleChanged(true); // trigger save options

					bool coreRootAtomHandled = false;

					// Update any active triggers
					foreach (KeyValuePair<Atom, AtomTriggerStates> pair in activeTriggersPerAtom)
					{
						pair.Value.AutoLoadTriggerRemoved(loadTrigger);

						// Force trigger unload on the newly saved trigger
						if (enableLoadTriggers.val && pair.Key != null && !pair.Key.destroyed)
						{
							coreRootAtomHandled |= (pair.Key == core.rootAtom);
							core.AutoLoaderTriggered(pair.Key, 2, null, new List<AutoLoadTriggerInfo>(new AutoLoadTriggerInfo[] { loadTrigger }));
						}
					}
					// UpdateAutoLoaderPresetsListForActiveTriggers(); // UpdateAutoLoaderPresetsList fixes this below

					if (!coreRootAtomHandled)
					{
						// Force trigger unload on the newly saved trigger
						if (enableLoadTriggers.val && core.rootAtom != null && !core.rootAtom.destroyed)
						{
							core.AutoLoaderTriggered(core.rootAtom, 2, null, new List<AutoLoadTriggerInfo>(new AutoLoadTriggerInfo[] { loadTrigger }));
						}
					}
				}
			}
			UpdateAutoLoaderPresetsList();
		}

		public List<string> GetPersonUidTriggerNames()
        {
			List<string> result = new List<string>();
			result.Add(""); // empty selection first!
			if (core.rootAtom != null && !core.rootAtom.destroyed)
            {
				result.Add(core.rootAtom.uid);
            }
			return result;
		}

		public const string TRIGGER_SORT_TRIGGER_DEFAULT = "Default";
		public const string TRIGGER_SORT_MORPH_NAME = "Morph Name";
		public const string TRIGGER_SORT_PRESET_NAME = "Preset Name";
		public const string TRIGGER_SORT_ACTIVE_TRIGGERS = "Active Triggers";
		public const string TRIGGER_SORT_PERSON_UID = "Person UiD";

		public List<string> GetTriggerSortingOptions()
        {
			List<string> result = new List<string>();
			result.Add(TRIGGER_SORT_TRIGGER_DEFAULT);
			result.Add(TRIGGER_SORT_MORPH_NAME);
			result.Add(TRIGGER_SORT_PRESET_NAME);
			result.Add(TRIGGER_SORT_ACTIVE_TRIGGERS);
			result.Add(TRIGGER_SORT_PERSON_UID);
			return result;
        }

		public List<string> GetMorphTriggerNames()
		{
			List<string> result = new List<string>();
			if (core.rootAtom != null && !core.rootAtom.destroyed)
			{
				DAZCharacterSelector geometry = core.rootAtom.GetStorableByID("geometry") as DAZCharacterSelector;
				if (geometry != null)
				{
					bool allowTriggerOnOtherGender = false;
					foreach (DAZMorph morph in geometry.morphBank1.morphs)
					{
						if (morph != null && morph.visible && morph.active && (morph.startValue != morph.appliedValue || morph.startValue != morph.morphValue))
						{
							if (morph.displayName == "CUAM_ReadOtherGenderMorphs")
							{
								allowTriggerOnOtherGender = true;
							} 
							else
                            {
								result.Add(morph.displayName);
							}
						}
					}
					if (allowTriggerOnOtherGender && geometry.morphBank1OtherGender != null)
                    {
						foreach (DAZMorph morph in geometry.morphBank1OtherGender.morphs)
						{
							if (morph != null && morph.visible && morph.active && (morph.startValue != morph.appliedValue || morph.startValue != morph.morphValue))
							{
								if (morph.displayName == "CUAM_ReadOtherGenderMorphs")
								{
									allowTriggerOnOtherGender = true;
								}
								else
								{
									result.Add(morph.displayName);
								}
							}
						}
					}
				}
			}
			result.Sort();
			result.Insert(0, ""); // empty selection first!
			result.Add(ANY_MORPH); // wildcard selection last (this will trigger on any person!)
			return result;
		}

		public List<string> GetLoadTriggerNames(bool includeEmpty)
		{
			List<string> triggerNames = new List<string>();
			if (includeEmpty)
			{
				triggerNames.Add("");
			}
			for (int i = 0; i < autoLoadTriggerPresets.Count; i++)
			{
				// build a short text that fits in the popup
				triggerNames.Add(string.Format("Trigger {0}", i.ToString()));
			}
			return triggerNames;
		}

		public bool CheckIncludeAtomType(string atomType)
        {
			// Checks if the atom should be included to be handled by this script.
			if (includedAtomTypes.Contains(atomType))
            {
				return true;
            }
			return false;
        }

		public List<string> GetIncludedAtomTypes()
        {
			return new List<string>(includedAtomTypes);
        }

		public List<string> GetPossibleLinkTypes()
        {
			return new List<string>(new string[]{
				FreeControllerV3.PositionState.On.ToString(),
				FreeControllerV3.PositionState.Comply.ToString(),
				FreeControllerV3.PositionState.Off.ToString(),
				FreeControllerV3.PositionState.ParentLink.ToString(),
				FreeControllerV3.PositionState.PhysicsLink.ToString(),
				FreeControllerV3.PositionState.Hold.ToString(),
				FreeControllerV3.PositionState.Lock.ToString()
			});
		}

		public string GetAtomLinkType(Atom apAtom)
        {
			if (apAtom.mainController.currentPositionState.ToString() == apAtom.mainController.currentRotationState.ToString())
            {
				return apAtom.mainController.currentPositionState.ToString();
			}
			// A mixed combination
			return apAtom.mainController.currentPositionState.ToString() + " / " + apAtom.mainController.currentRotationState.ToString();
		}

		public FreeControllerV3.PositionState GetAtomLinkTypePosition(string linkType, FreeControllerV3.PositionState defaultValue)
        {
			switch(linkType)
            {
				case "On": return FreeControllerV3.PositionState.On;
				case "Comply": return FreeControllerV3.PositionState.Comply;
				case "Off": return FreeControllerV3.PositionState.Off;
				case "ParentLink": return FreeControllerV3.PositionState.ParentLink;
				case "PhysicsLink": return FreeControllerV3.PositionState.PhysicsLink;
				case "Hold": return FreeControllerV3.PositionState.Hold;
				case "Lock": return FreeControllerV3.PositionState.Lock;
			}
			return defaultValue;
		}

		public FreeControllerV3.RotationState GetAtomLinkTypeRotation(string linkType, FreeControllerV3.RotationState defaultValue)
		{
			switch (linkType)
			{
				case "On": return FreeControllerV3.RotationState.On;
				case "Comply": return FreeControllerV3.RotationState.Comply;
				case "Off": return FreeControllerV3.RotationState.Off;
				case "ParentLink": return FreeControllerV3.RotationState.ParentLink;
				case "PhysicsLink": return FreeControllerV3.RotationState.PhysicsLink;
				case "Hold": return FreeControllerV3.RotationState.Hold;
				case "Lock": return FreeControllerV3.RotationState.Lock;
			}
			return defaultValue;
		}

		public bool TrackLinkInWorldSpace(FreeControllerV3.PositionState state)
        {
			return false; // (state != FreeControllerV3.PositionState.ParentLink && state != FreeControllerV3.PositionState.PhysicsLink);
		}

		public bool TrackLinkInWorldSpace(FreeControllerV3.RotationState state)
		{
			return false; // (state != FreeControllerV3.RotationState.ParentLink && state != FreeControllerV3.RotationState.PhysicsLink);
		}

		public bool LiveUpdateLink(FreeControllerV3.PositionState state)
		{
			return (state != FreeControllerV3.PositionState.ParentLink && state != FreeControllerV3.PositionState.PhysicsLink);
		}

		public bool LiveUpdateLink(FreeControllerV3.RotationState state)
		{
			return (state != FreeControllerV3.RotationState.ParentLink && state != FreeControllerV3.RotationState.PhysicsLink);
		}


	}

	public class AtomTriggerStates
    {
		public List<AutoLoadTriggerInfo> activeTriggers;

		public int lastMorphBankCount = 0;
		public int lastAutoLoadTriggerMorphNamesCount = 0;
		public DAZCharacterSelector.Gender gender;
		public bool dirty;
		public List<DAZMorph> useOtherGenderMorphs;

		public int lastMorphOtherGenderBankCount = 0;
		public bool lastUseOtherGenderMorph = false;


		/// <summary>
		/// Morph bank containing morph displayName and all morphs using that displayName. This plugin as a whole operates on morph displayNames...
		/// </summary>
		public Dictionary<string, List<DAZMorph>> morphBank;

		public AtomTriggerStates()
        {
			gender = DAZCharacterSelector.Gender.None;
			activeTriggers = new List<AutoLoadTriggerInfo>();
			lastMorphBankCount = 0;
			morphBank = null;
			dirty = false;
			useOtherGenderMorphs = null;
			lastUseOtherGenderMorph = false;

			lastMorphOtherGenderBankCount = 0;
		}

		public void AutoLoadTriggerAdded(AutoLoadTriggerInfo loadTrigger)
        {
			dirty = true;
		}

		public void AutoLoadTriggerRemoved(AutoLoadTriggerInfo loadTrigger)
        {
			// AutoLoadTrigger removed, no longer active
			// activeTriggers.Remove(loadTrigger); !?
			dirty = true;
		}

		/// <summary>
        /// Gets the state morph bank (or the cached version).
        /// This assumes morphs are added to the morphBank as they are found from .var packages as they are loaded into memory and not removed.
        /// </summary>
        /// <param name="geometry"></param>
        /// <param name="autoLoadTriggerMorphNames"></param>
        /// <returns></returns>
		public Dictionary<string, List<DAZMorph>> GetStateMorphBank(DAZCharacterSelector geometry, HashSet<string> autoLoadTriggerMorphNames)
        {
			int morphBankCount = geometry.morphBank1.morphs.Count;
			bool useOtherGenderMorphBank = false;

			// if (geometry.gender == )
			// geometry.morphBank1OtherGender.morphs.Count
			// useFemaleMorphsOnMale
			// useMaleMorphsOnFemale

			if (morphBank == null || dirty || gender != geometry.gender || lastMorphBankCount != geometry.morphBank1.morphs.Count || lastAutoLoadTriggerMorphNamesCount != autoLoadTriggerMorphNames.Count)
            {
				dirty = false;
				morphBank = new Dictionary<string, List<DAZMorph>>();
				gender = geometry.gender;
				lastMorphBankCount = geometry.morphBank1.morphs.Count;
				lastAutoLoadTriggerMorphNamesCount = autoLoadTriggerMorphNames.Count;
				
				lastMorphOtherGenderBankCount = 0;
				lastUseOtherGenderMorph = false;
				useOtherGenderMorphs = new List<DAZMorph>();

				List<string> allAddedMorphs = null;
				if (CustomUnityAssetManager.DEBUG) 
					allAddedMorphs = new List<string>();

				List<DAZMorph> morphs = null;
				foreach (DAZMorph morph in geometry.morphBank1.morphs.Where((morph) => { 
					return autoLoadTriggerMorphNames.Contains(morph.displayName) || morph.displayName == "CUAM_ReadOtherGenderMorphs"; 
				}))
				{
					if (morph.displayName == "CUAM_ReadOtherGenderMorphs")
					{
						useOtherGenderMorphs.Add(morph);
					}
					else
					{
						if (!morphBank.TryGetValue(morph.displayName, out morphs))
						{
							morphs = new List<DAZMorph>();
							morphBank.Add(morph.displayName, morphs);
						}
						morphs.Add(morph);

						if (CustomUnityAssetManager.DEBUG)
							allAddedMorphs.Add(morph.displayName);
					}
				}

				if (CustomUnityAssetManager.DEBUG)
					SuperController.LogMessage("MorphBank rebuilt"); // with: " + string.Join(", ", allAddedMorphs.OrderBy(x => x).ToArray()));
				
			}

			if (useOtherGenderMorphs != null && useOtherGenderMorphs.Any((x) => { return x.visible && x.active && (x.startValue != x.appliedValue || x.startValue != x.morphValue); }))
			{
				useOtherGenderMorphBank = true;
			}

			if (geometry.morphBank1OtherGender != null && (useOtherGenderMorphBank || lastUseOtherGenderMorph))
			{
				if (lastMorphOtherGenderBankCount != geometry.morphBank1OtherGender.morphs.Count || lastUseOtherGenderMorph != useOtherGenderMorphBank)
				{
					lastMorphOtherGenderBankCount = geometry.morphBank1OtherGender.morphs.Count;
					lastUseOtherGenderMorph = useOtherGenderMorphBank;

					List<string> allAddedMorphs = null;
					List<DAZMorph> morphs = null;

					if (useOtherGenderMorphBank)
					{
						foreach (DAZMorph morph in geometry.morphBank1OtherGender.morphs.Where((morph) => { return autoLoadTriggerMorphNames.Contains(morph.displayName); }))
						{
							if (!morphBank.TryGetValue(morph.displayName, out morphs))
							{
								morphs = new List<DAZMorph>();
								morphBank.Add(morph.displayName, morphs);
							}
							morphs.Add(morph);

							if (CustomUnityAssetManager.DEBUG)
								allAddedMorphs.Add(morph.displayName);
						}
					} 
					else
                    {
						foreach (DAZMorph morph in geometry.morphBank1OtherGender.morphs.Where((morph) => { return autoLoadTriggerMorphNames.Contains(morph.displayName); }))
						{
							if (morphBank.TryGetValue(morph.displayName, out morphs))
							{
								morphs.Remove(morph);
							}
						}
					}

					if (CustomUnityAssetManager.DEBUG)
						SuperController.LogMessage("MorphBankOtherGender rebuilt"); // with: " + string.Join(", ", allAddedMorphs.OrderBy(x => x).ToArray()));
				}
			}


			return morphBank;
		}

	}

	// wrapper class for the list info to make life easier
	public class AutoLoadTriggerListInfo
    {
		public int Index;
		public AutoLoadTriggerInfo Info;
		public AutoLoadTriggerListInfo(AutoLoadTriggerInfo info, int index)
        {
			this.Info = info;
			this.Index = index;
		}
	}

	public class AutoLoadTriggerInfo
    {
		public string personUid;
		public string morphName;
		public string presetName;
		public string presetType;
		public bool external;
		public string externalFile;

		public AutoLoadTriggerInfo(string uid, string morph, string preset, string extFile, string type)
        {
			personUid = (uid == null ? "" : uid);
			morphName = (morph == null ? "" : morph);
			presetName = (preset == null ? "" : preset);
			presetType = (type == null ? "" : type);
			if (presetType == "")
				presetType = Options.PRESET_CUAM; // assuming CUAM as default if nothing else is provided
			externalFile = (extFile == null ? "" : extFile);
			external = (externalFile != "");
		}

		public bool Equals(AutoLoadTriggerInfo other)
        {
			if (other == null)
				return false;

			return (personUid == other.personUid && 
					morphName == other.morphName && 
					presetName == other.presetName && 
					presetType == other.presetType &&
					externalFile == other.externalFile && 
					external == other.external);
        }

        public override string ToString()
        {
			return string.Format("{0}:{1}:{2}:{3}:{4}:{5}", personUid, morphName, presetName, presetType, external.ToString(), externalFile);
        }

    }

}