using System;
using UnityEngine;
using UnityEngine.EventSystems;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Blazedust.CUAM
{
	public class PosRotSliders
	{

		private JSONStorableFloat[] positionOffsets;
		private JSONStorableFloat[] rotationOffsets;

		private UIDynamicSlider[] positionOffsetSliders;
		private UIDynamicSlider[] rotationOffsetSliders;

		private bool positionIsOffset;
		private bool rotationIsOffset;

		private CustomUnityAssetManager core;

		public PosRotSliders(CustomUnityAssetManager script)
        {
			core = script;
		}

		public void InitStorables(JSONStorableFloat.SetJSONFloatCallback sliderChanged)
        {
			positionOffsets = new JSONStorableFloat[3];
			positionOffsetSliders = new UIDynamicSlider[3];
			positionOffsets[0] = new JSONStorableFloat("Position X offset", 0.0f, sliderChanged, -2f, 2f, false);
			positionOffsets[1] = new JSONStorableFloat("Position Y offset", 0.0f, sliderChanged, -2f, 2f, false);
			positionOffsets[2] = new JSONStorableFloat("Position Z offset", 0.0f, sliderChanged, -2f, 2f, false);


			// X,Y,Z offset sliders for rotation
			rotationOffsets = new JSONStorableFloat[3];
			rotationOffsetSliders = new UIDynamicSlider[3];
			rotationOffsets[0] = new JSONStorableFloat("Rotation X offset", 0.0f, sliderChanged, -360f, 360f, true);
			rotationOffsets[1] = new JSONStorableFloat("Rotation Y offset", 0.0f, sliderChanged, -360f, 360f, true);
			rotationOffsets[2] = new JSONStorableFloat("Rotation Z offset", 0.0f, sliderChanged, -360f, 360f, true);
		}

		public void CreateGUI(UIHelper uiHelper)
        {
			positionOffsetSliders[0] = uiHelper.CreateSlider(positionOffsets[0]);
			positionOffsetSliders[0].valueFormat = "F3";
			positionOffsetSliders[0].ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);

			positionOffsetSliders[1] = uiHelper.CreateSlider(positionOffsets[1]);
			positionOffsetSliders[1].valueFormat = "F3";
			positionOffsetSliders[1].ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);

			positionOffsetSliders[2] = uiHelper.CreateSlider(positionOffsets[2]);
			positionOffsetSliders[2].valueFormat = "F3";
			positionOffsetSliders[2].ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);

			// X,Y,Z offset sliders for rotation
			rotationOffsetSliders[0] = uiHelper.CreateSlider(rotationOffsets[0], true);
			rotationOffsetSliders[0].ConfigureQuickButtons(-1f, -5f, -10f, -45f, 1f, 5f, 10f, 45f);

			rotationOffsetSliders[1] = uiHelper.CreateSlider(rotationOffsets[1], true);
			rotationOffsetSliders[1].ConfigureQuickButtons(-1f, -5f, -10f, -45f, 1f, 5f, 10f, 45f);

			rotationOffsetSliders[2] = uiHelper.CreateSlider(rotationOffsets[2], true);
			rotationOffsetSliders[2].ConfigureQuickButtons(-1f, -5f, -10f, -45f, 1f, 5f, 10f, 45f);

			UpdateSliderLabels();
		}

		public void GUIDestroyed()
        {
			positionOffsetSliders[0] = null;
			positionOffsetSliders[1] = null;
			positionOffsetSliders[2] = null;
			rotationOffsetSliders[0] = null;
			rotationOffsetSliders[1] = null;
			rotationOffsetSliders[2] = null;

		}

		private void UpdateSliderLabels()
        {
			if (!positionIsOffset)
			{
				if (positionOffsetSliders[0] != null)
				{
					positionOffsetSliders[0].labelText.text = "Position X";
					positionOffsetSliders[1].labelText.text = "Position Y";
					positionOffsetSliders[2].labelText.text = "Position Z";
				}
			}
			else
			{
				if (positionOffsetSliders[0] != null)
				{
					positionOffsetSliders[0].labelText.text = "Position X offset";
					positionOffsetSliders[1].labelText.text = "Position Y offset";
					positionOffsetSliders[2].labelText.text = "Position Z offset";
				}
			}
			if (!rotationIsOffset)
			{
				if (rotationOffsetSliders[0] != null)
				{
					rotationOffsetSliders[0].labelText.text = "Rotation X";
					rotationOffsetSliders[1].labelText.text = "Rotation Y";
					rotationOffsetSliders[2].labelText.text = "Rotation Z";
				}
			}
			else
			{
				if (rotationOffsetSliders[0] != null)
				{
					rotationOffsetSliders[0].labelText.text = "Rotation X offset";
					rotationOffsetSliders[1].labelText.text = "Rotation Y offset";
					rotationOffsetSliders[2].labelText.text = "Rotation Z offset";
				}
			}
		}

		public void Destroy()
        {
			core = null;
		}

		/// <summary>
		/// Updates the plugin offset pos/rot values in the plugin GUI from selected atom/link.
		/// </summary>
		public void UpdateOffsetsPosRotFromAtom(Atom apAtom, Atom parentAtom, Rigidbody rbLink, bool readPosition = true, bool readRotation = true)
		{
			Vector3 linkPos = Vector3.zero;
			Vector3 linkRot = Vector3.zero;
			if (rbLink != null)
			{
				linkPos = rbLink.transform.InverseTransformPoint(apAtom.childAtomContainer.position);
				linkRot = (Quaternion.Inverse(apAtom.childAtomContainer.transform.rotation) * rbLink.transform.rotation).eulerAngles;

				if (readPosition)
				{
					if (core.options.TrackLinkInWorldSpace(apAtom.mainController.currentPositionState))
					{
						linkPos = (apAtom.childAtomContainer != null ? apAtom.childAtomContainer.transform.position : apAtom.transform.position);
						positionIsOffset = false;
					}
					else
					{
						positionIsOffset = true;
					}
				}
				if (readRotation)
				{
					if (core.options.TrackLinkInWorldSpace(apAtom.mainController.currentRotationState))
					{
						linkRot = (apAtom.childAtomContainer != null ? apAtom.childAtomContainer.transform.rotation : apAtom.transform.rotation).eulerAngles;
						rotationIsOffset = false;
					}
					else
					{
						rotationIsOffset = true;
					}
				}
				if (readPosition || readRotation)
				{
					UpdateSliderLabels();
				}
			}
			if (readPosition)
			{
				positionOffsets[0].valNoCallback = linkPos.x;
				positionOffsets[1].valNoCallback = linkPos.y;
				positionOffsets[2].valNoCallback = linkPos.z;
			}
			if (readRotation)
			{
				rotationOffsets[0].valNoCallback = linkRot.x;
				rotationOffsets[1].valNoCallback = linkRot.y;
				rotationOffsets[2].valNoCallback = linkRot.z;
			}
		}

		public Vector3 Position
        {
			get
            {
				return new Vector3(positionOffsets[0].valNoCallback, positionOffsets[1].valNoCallback, positionOffsets[2].valNoCallback);
            }
			set
            {
				positionOffsets[0].valNoCallback = value.x;
				positionOffsets[1].valNoCallback = value.y;
				positionOffsets[2].valNoCallback = value.z;
			}
        }

		public Vector3 Rotation
		{
			get
			{
				return new Vector3(rotationOffsets[0].valNoCallback, rotationOffsets[1].valNoCallback, rotationOffsets[2].valNoCallback);
			}
			set
			{
				rotationOffsets[0].valNoCallback = value.x;
				rotationOffsets[1].valNoCallback = value.y;
				rotationOffsets[2].valNoCallback = value.z;
			}
		}

	}
}