using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Blazedust.CUAM
{
	/// <summary>
	/// Helper to track CustomUnityAssets belonging to the scene.
	/// </summary>
	public class SceneAtomTracker
	{
		public List<Atom> sceneCuas;

		private int trackerPos = 0;
		private bool resetTracker = true;
		private CustomUnityAssetManager core;

		public SceneAtomTracker(CustomUnityAssetManager script)
		{
			core = script;
			sceneCuas = new List<Atom>();
			resetTracker = true;
		}

		public void Destroy()
        {
			core = null;
			sceneCuas.Clear();
			sceneCuas = null;
		}

		public void Reset()
        {
			resetTracker = true;
		}

		/// <summary>
		/// Keeps track on scene CustomUnityAssets
		/// </summary>
		public void Update()
		{
			if (resetTracker)
			{
				// SuperController.LogMessage("Updating Scene CUAs...");
				List<Atom> atoms = SuperController.singleton.GetAtoms();
				sceneCuas.Clear();
				sceneCuas.AddRange(atoms.Where((x) => { return core.CheckIncludeAtomType(x.type); }));
				// SuperController.LogMessage("Scene Cuas: " + string.Join(",", sceneCuas.Select((x) => { return x.uid; }).ToArray()));
				resetTracker = false;
			}
			else if (sceneCuas.Count > 0)
			{
				// Round robin destroyed check
				trackerPos = (trackerPos + 1) % sceneCuas.Count;
				if (sceneCuas[trackerPos].destroyed)
				{
					while (sceneCuas.Count > 0 && sceneCuas[trackerPos].destroyed)
					{
						sceneCuas.RemoveAt(trackerPos);
						if (sceneCuas.Count == 0)
						{
							break;
						}
						trackerPos = (trackerPos + 1) % sceneCuas.Count;
					}
					// SuperController.LogMessage("Scene Cuas (after some destroyed): " + string.Join(",", sceneCuas.Select((x) => { return x.uid; }).ToArray()));
				}
			}
		}
	}
}