using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;

namespace Blazedust.CUAM
{

    public class UIGroupBrowser
    {
        UIHelper uiHelper;

        UIGroup groupHome;
        List<UIGroup> selectableGroups;
        public string activeGroupKey { get; private set; }
        string groupHomeKey;

        JSONStorableStringChooser groupSelector;

        public UIGroupBrowser(MVRScript script)
        {
            uiHelper = new UIHelper(script);
            selectableGroups = new List<UIGroup>();
            activeGroupKey = "";
        }

        public UIDynamicButton CreateHeader()
        {
            UIDynamicButton btn = uiHelper.CreateButton("Home");
            // btn.buttonColor = Color.yellow;
            btn.button.onClick.AddListener(() =>
            {
                try
                {
                    if (activeGroupKey != groupHomeKey)
                    {
                        ActivateGroup(groupHomeKey);
                    }
                }
                catch (Exception e)
                {
                    SuperController.LogError(e.ToString());
                }
            });

            groupSelector = new JSONStorableStringChooser("cuabGroupSelector", GetGroupKeys(), "", "Goto Page:", (string key) => {
                if (activeGroupKey != key)
                {
                    try
                    {
                        ActivateGroup(key);
                    }
                    catch (Exception e)
                    {
                        SuperController.LogError(e.ToString());
                    }
                }
            });
            UIDynamicPopup popup = uiHelper.CreatePopup(groupSelector, true);
            popup.popupPanelHeight = 400f;
            groupSelector.popup.onOpenPopupHandlers += () => {
                groupSelector.choices = GetGroupKeys();
            };

            btn.height = popup.height;

            return btn;
            //btn.height = 80;
            //uiHelper.CreateSpacer(false, (int)(popup.height - btn.height - 20
        }

        private List<string> GetGroupKeys()
        {
            List<string> keys = new List<string>();
            foreach (UIGroup group in selectableGroups)
                keys.Add(group.Key);
            return keys;
        }

        public bool ActivateGroup(string key)
        {
            foreach (UIGroup group in selectableGroups)
            {
                if (group.Key == key)
                {
                    if (false)
                    {
                        // Create, Hide/Show groups. If Hide/Show can work without problems that is...
                        HideActiveGroup();
                        activeGroupKey = group.Key;
                        groupSelector.valNoCallback = activeGroupKey;
                        if (group.Created)
                            group.ShowGUI();
                        else
                            group.CreateGUI();
                    } 
                    else
                    {
                        RemoveActiveGroups();
                        activeGroupKey = group.Key;
                        groupSelector.valNoCallback = activeGroupKey;
                        group.CreateGUI();
                    }
                    
                    return true;
                }
            }
            activeGroupKey = "";
            return false;
        }

        public void SetHome(UIGroup group)
        {
            groupHome = group;
            groupHomeKey = (group != null ? group.Key : "");
        }

        public void AddGroup(UIGroup group)
        {
            selectableGroups.Add(group);
            groupSelector.choices = GetGroupKeys();
        }

        /// <summary>
        /// Creates a group and adds it to the list of available groups. 
        /// This sets the home group if it's not yet set.
        /// This actives the group if no group yet is activate.
        /// </summary>
        /// <param name="key"></param>
        /// <param name="createGUICallback"></param>
        /// <returns></returns>
        public UIGroup CreateGroup(string key, Action<UIHelper> createGUICallback, Action<UIHelper> removeGUICallback = null)
        {
            UIGroup group = new UIGroup(uiHelper.core, key, createGUICallback, removeGUICallback);
            if (groupHome == null)
                SetHome(group);
            AddGroup(group);
            if (activeGroupKey == "")
                ActivateGroup(key);
            return group;
        }

        private void HideActiveGroup()
        {
            foreach (UIGroup group in selectableGroups)
            {
                if (group.Created && !group.Hidden)
                {
                    group.HideGUI();
                }
            }
        }

        private void RemoveActiveGroups()
        {
            foreach (UIGroup group in selectableGroups)
            {
                group.RemoveGUI();
            }
        }

        public void Destroy()
        {
            RemoveActiveGroups();
            selectableGroups.Clear();
            groupHome = null;
            uiHelper.Destroy();
        }
    }

    public class UIGroup
    {

        public string Key;
        
        UIHelper uiHelper;

        Action<UIHelper> createCallback;
        Action<UIHelper> removeCallback;

        bool hidden = false;

        public UIGroup(MVRScript script, string key, Action<UIHelper> createGUICallback, Action<UIHelper> removeGUICallback = null)
        {
            this.Key = key;
            this.createCallback = createGUICallback;
            this.removeCallback = removeGUICallback;
            this.uiHelper = new UIHelper(script);
        }

        public void CreateGUI()
        {
            RemoveGUI();
            this.createCallback.Invoke(uiHelper);
        }

        public bool Created { get { return uiHelper.controls.Count > 0; } }

        public bool Hidden { get { return hidden; } }

        public void HideGUI()
        {
            hidden = true;
            uiHelper.SetEnabled(false);
        }
        public void ShowGUI()
        {
            hidden = false;
            uiHelper.SetEnabled(true);
        }

        public void RemoveGUI()
        {
            if (uiHelper.controls.Count > 0)
            {
                if (removeCallback != null)
                {
                    removeCallback.Invoke(uiHelper);
                }
                uiHelper.Destroy();
            }
        }

    }
}